from typing import Dict, Union

import aim.ext.transport.remote_tracking_pb2 as tracking_rpc
import aim.ext.transport.remote_tracking_pb2_grpc as tracking_pb2_grpc
import aim.ext.transport.message_utils as utils

from aim.ext.transport.handlers import get_handler
from aim.storage.treeutils import encode_tree, decode_tree


class ResourceTypeRegistry:
    def __init__(self):
        self._registry: Dict[str, type] = {}

    def register(self, type_name: str, resource_getter: Union[type, callable]):
        self._registry[type_name] = resource_getter

    def __getitem__(self, type_name: str):
        return self._registry[type_name]


class RemoteTrackingServicer(tracking_pb2_grpc.RemoteTrackingServiceServicer):
    resource_pool = dict()

    def __init__(self, resource_registry: ResourceTypeRegistry):
        self.registry = resource_registry

    def cleanup_client_resources(self, request: tracking_rpc.ClientResourceCleanupRequest, _context):
        dead_client_uri = request.client_uri

        resource_handlers = list(self.resource_pool.keys())
        for handler in resource_handlers:
            (client_uri, _) = self.resource_pool[handler]
            if dead_client_uri == client_uri:
                del self.resource_pool[handler]

        return tracking_rpc.ResourceResponse(status=tracking_rpc.ResourceResponse.Status.OK)

    def get_resource(self, request: tracking_rpc.ResourceRequest, _context):
        if not request.handler:
            resource_handler = get_handler()
        else:
            resource_handler = request.handler

        try:
            resource_cls = self.registry[request.resource_type]
            if len(request.args) > 0:
                res = resource_cls(request.args)
            else:
                res = resource_cls()

            self.resource_pool[resource_handler] = (request.client_uri, res)
            return tracking_rpc.ResourceResponse(handler=resource_handler,
                                                 status=tracking_rpc.ResourceResponse.Status.OK)
        except Exception as e:
            try:
                # clean up resource_pool
                # because no one will call release_resource in case of Exception
                del self.resource_pool[resource_handler]
            except KeyError:
                pass

            return tracking_rpc.ResourceResponse(
                handler=None,
                status=tracking_rpc.ResourceResponse.Status.ERROR,
                exception=utils.build_exception(e),
            )

    def release_resource(self, request: tracking_rpc.ReleaseResourceRequest, _context):
        try:
            resource_handler = request.handler
            client_uri = request.client_uri
            self._verify_resource_handler(resource_handler, client_uri)
            del self.resource_pool[resource_handler]
            return tracking_rpc.ReleaseResourceResponse(status=tracking_rpc.ReleaseResourceResponse.Status.OK)
        except Exception as e:
            # Should we clean ed resource_pool in case of exceptions? what if UnauthorizedRequest?
            # del self.resource_pool[resource_handler]
            return tracking_rpc.ReleaseResourceResponse(
                status=tracking_rpc.ReleaseResourceResponse.Status.ERROR,
                exception=utils.build_exception(e),
            )

    def run_instruction(self, request_iterator, _context) -> tracking_rpc.InstructionResponse:
        try:
            header = next(request_iterator)
            assert header.WhichOneof('instruction') == 'header'

            resource_handler = header.header.handler
            client_uri = header.header.client_uri
            self._verify_resource_handler(resource_handler, client_uri)

            args = decode_tree(utils.unpack_stream(request_iterator))
            checked_args = []
            for arg in args:
                if isinstance(arg, utils.ResourceObject):
                    handler = arg.storage['handler']
                    self._verify_resource_handler(handler, client_uri)
                    checked_args.append(self.resource_pool[handler][1].ref)
                else:
                    checked_args.append(arg)

            method_name = header.header.method_name

            resource = self.resource_pool[resource_handler][1].ref
            if method_name.endswith('.setter'):
                attr_name = method_name.split('.')[0]
                setattr(resource, attr_name, checked_args[0])
                result = None
            else:
                attr = getattr(resource, method_name)
                if callable(attr):
                    result = attr(*checked_args)
                else:
                    result = attr

            yield tracking_rpc.InstructionResponse(header=tracking_rpc.ResponseHeader(
                version='0.1',
                status=tracking_rpc.ResponseHeader.Status.OK
            ))
            for chunk in utils.pack_stream(encode_tree(result)):
                yield tracking_rpc.InstructionResponse(message=chunk)

        except Exception as e:
            yield tracking_rpc.InstructionResponse(header=tracking_rpc.ResponseHeader(
                version='0.1',
                status=tracking_rpc.ResponseHeader.Status.ERROR,
                exception=utils.build_exception(e),
            ))
            return

    def run_write_instructions(self, request_iterator, _context) -> tracking_rpc.WriteInstructionsResponse:
        try:
            raw_message = []
            client_uri = None  # TODO [AD] move to header interface?
            # TODO [AD] raw_message = [request.message for request in request_iterator]
            for request in request_iterator:
                raw_message.append(request.message)
                client_uri = request.client_uri
            write_instructions = decode_tree(utils.unpack_bytes(raw_message))
            for instruction in write_instructions:
                resource_handler, method_name, args = instruction

                self._verify_resource_handler(resource_handler, client_uri)

                checked_args = []
                for arg in args:
                    if isinstance(arg, utils.ResourceObject):
                        handler = arg.storage['handler']
                        self._verify_resource_handler(handler, client_uri)
                        checked_args.append(self.resource_pool[handler][1].ref)
                    else:
                        checked_args.append(arg)

                resource = self.resource_pool[resource_handler][1].ref
                if method_name.endswith('.setter'):
                    attr_name = method_name.split('.')[0]
                    setattr(resource, attr_name, checked_args[0])
                else:
                    attr = getattr(resource, method_name)
                    assert callable(attr)
                    attr(*checked_args)

            return tracking_rpc.WriteInstructionsResponse(status=tracking_rpc.WriteInstructionsResponse.Status.OK)
        except Exception as e:
            return tracking_rpc.WriteInstructionsResponse(
                version='0.1',
                status=tracking_rpc.WriteInstructionsResponse.Status.ERROR,
                exception=utils.build_exception(e),
            )

    def _verify_resource_handler(self, resource_handler, client_uri):
        res_info = self.resource_pool.get(resource_handler, None)
        if not res_info or res_info[0] != client_uri:
            raise utils.UnauthorizedRequestError(resource_handler)



from aim.sdk.objects.audio import Audio
from aim.sdk.objects.image import Image
from aim.sdk.objects.figure import Figure
from aim.sdk.objects.distribution import Distribution
from aim.sdk.objects.text import Text



import cython
import pyximport

import logging
logger = logging.getLogger(__name__)


# if not cython.compiled:
pyximport.install(inplace=True,
                  language_level=3,
                  build_in_temp=False,
                  setup_args={'language': 'c++'})


def log_status():
    logger.info(f'compiled: {cython.compiled}')



import os

AIM_ENABLE_TRACKING_THREAD = '__AIM_ENABLE_TRACKING_THREAD__'
AIM_REPO_NAME = '__AIM_REPO_NAME__'
AIM_RUN_INDEXING_TIMEOUT = '__AIM_RUN_INDEXING_TIMEOUT_SECONDS__'


def get_aim_repo_name():
    return os.environ.get(AIM_REPO_NAME) or '.aim'



from fastapi import HTTPException, Depends
from aim.web.api.utils import APIRouter  # wrapper for fastapi.APIRouter
from typing import Optional

from aim.web.api.projects.project import Project
from aim.web.api.utils import object_factory
from aim.web.api.tags.pydantic_models import (
    TagCreateIn,
    TagUpdateIn,
    TagUpdateOut,
    TagGetOut,
    TagListOut,
    TagGetRunsOut,
)
tags_router = APIRouter()


@tags_router.get('/', response_model=TagListOut)
async def get_tags_list_api(factory=Depends(object_factory)):
    return [{'id': tag.uuid,
             'name': tag.name,
             'color': tag.color,
             'description': tag.description,
             'run_count': len(tag.runs),
             'archived': tag.archived}
            for tag in factory.tags()]


@tags_router.get('/search/', response_model=TagListOut)
async def search_tags_by_name_api(q: Optional[str] = '', factory=Depends(object_factory)):
    return [{'id': tag.uuid,
             'name': tag.name,
             'color': tag.color,
             'description'
             'run_count': len(tag.runs),
             'archived': tag.archived}
            for tag in factory.search_tags(q.strip())]


@tags_router.post('/', response_model=TagUpdateOut)
async def create_tag_api(tag_in: TagCreateIn, factory=Depends(object_factory)):
    with factory:
        try:
            tag = factory.create_tag(tag_in.name.strip())
            if tag_in.color is not None:
                tag.color = tag_in.color.strip()
            if tag_in.description is not None:
                tag.description = tag_in.description.strip()
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e))

    return {
        'id': tag.uuid,
        'status': 'OK'
    }


@tags_router.get('/{tag_id}/', response_model=TagGetOut)
async def get_tag_api(tag_id: str, factory=Depends(object_factory)):
    tag = factory.find_tag(tag_id)
    if not tag:
        raise HTTPException(status_code=404)

    response = {
        'id': tag.uuid,
        'name': tag.name,
        'color': tag.color,
        'description': tag.description,
        'archived': tag.archived,
        'run_count': len(tag.runs)
    }
    return response


@tags_router.put('/{tag_id}/', response_model=TagUpdateOut)
async def update_tag_properties_api(tag_id: str, tag_in: TagUpdateIn, factory=Depends(object_factory)):
    with factory:
        tag = factory.find_tag(tag_id)
        if not tag:
            raise HTTPException(status_code=404)

        if tag_in.name:
            tag.name = tag_in.name.strip()
        if tag_in.color is not None:
            tag.color = tag_in.color.strip()
        if tag_in.description is not None:
            tag.description = tag_in.description.strip()
        if tag_in.archived is not None:
            tag.archived = tag_in.archived

    return {
        'id': tag.uuid,
        'status': 'OK'
    }


@tags_router.delete('/{tag_id}/')
async def delete_tag_api(tag_id: str, factory=Depends(object_factory)):
    with factory:
        if not factory.delete_tag(tag_id):
            raise HTTPException(status_code=404)


@tags_router.get('/{tag_id}/runs/', response_model=TagGetRunsOut)
async def get_tagged_runs_api(tag_id: str, factory=Depends(object_factory)):
    project = Project()

    tag = factory.find_tag(tag_id)
    if not tag:
        raise HTTPException(status_code=404)

    from aim.sdk.run import Run

    cache_name = 'tag_runs'
    project.repo.run_props_cache_hint = cache_name
    project.repo.structured_db.invalidate_cache(cache_name)
    project.repo.structured_db.init_cache(cache_name, tag.get_runs, lambda run_: run_.hash)

    tag_runs = []
    for tagged_run in tag.runs:
        run = Run(tagged_run.hash, repo=project.repo, read_only=True)
        tag_runs.append({
            'run_id': tagged_run.hash,
            'name': tagged_run.name,
            'creation_time': run.creation_time,
            'end_time': run.end_time,
            'experiment': tagged_run.experiment if tagged_run.experiment else None
        })

    project.repo.structured_db.invalidate_cache(cache_name)
    project.repo.run_props_cache_hint = None

    response = {
        'id': tag.uuid,
        'runs': tag_runs
    }
    return response



from fastapi import Depends, HTTPException
from aim.web.api.utils import APIRouter  # wrapper for fastapi.APIRouter
from sqlalchemy.orm import Session
from typing import List
from aim.web.api.dashboards.models import Dashboard
from aim.web.api.dashboard_apps.models import ExploreState
from aim.web.api.dashboards.serializers import dashboard_response_serializer
from aim.web.api.db import get_session
from aim.web.api.dashboards.pydantic_models import (
    DashboardOut,
    DashboardCreateIn,
    DashboardUpdateIn
)

dashboards_router = APIRouter()


@dashboards_router.get('/', response_model=List[DashboardOut])
async def dashboards_list_api(session: Session = Depends(get_session)):
    dashboards_query = session.query(Dashboard) \
        .filter(Dashboard.is_archived == False) \
        .order_by(Dashboard.updated_at)  # noqa
    result = []

    for dashboard in dashboards_query:
        result.append(dashboard_response_serializer(dashboard, session))

    return result


@dashboards_router.post('/', status_code=201, response_model=DashboardOut)
async def dashboards_post_api(request_data: DashboardCreateIn, session: Session = Depends(get_session)):
    # create the dashboard object
    dashboard_name = request_data.name
    dashboard_description = request_data.description
    dashboard = Dashboard(dashboard_name, dashboard_description)
    session.add(dashboard)

    # update the app object's foreign key relation
    app_id = str(request_data.app_id)
    app = session.query(ExploreState).filter(ExploreState.uuid == app_id).first()
    if app:
        app.dashboard_id = dashboard.uuid

    # commit db session
    session.commit()

    return dashboard_response_serializer(dashboard, session)


@dashboards_router.get('/{dashboard_id}/', response_model=DashboardOut)
async def dashboards_get_api(dashboard_id: str, session: Session = Depends(get_session)):
    dashboard = session.query(Dashboard) \
        .filter(Dashboard.uuid == dashboard_id, Dashboard.is_archived == False) \
        .first()  # noqa
    if not dashboard:
        raise HTTPException(status_code=404)

    return dashboard_response_serializer(dashboard, session)


@dashboards_router.put('/{dashboard_id}/', response_model=DashboardOut)
async def dashboards_put_api(dashboard_id: str, request_data: DashboardUpdateIn,
                             session: Session = Depends(get_session)):
    dashboard = session.query(Dashboard) \
        .filter(Dashboard.uuid == dashboard_id, Dashboard.is_archived == False) \
        .first()  # noqa
    if not dashboard:
        raise HTTPException(status_code=404)
    dashboard_name = request_data.name
    if dashboard_name:
        dashboard.name = dashboard_name
    dashboard_description = request_data.description
    if dashboard_description:
        dashboard.description = dashboard_description
    session.commit()

    return dashboard_response_serializer(dashboard, session)


@dashboards_router.delete('/{dashboard_id}/')
async def dashboards_delete_api(dashboard_id: str, session: Session = Depends(get_session)):
    dashboard = session.query(Dashboard) \
        .filter(Dashboard.uuid == dashboard_id, Dashboard.is_archived == False) \
        .first()  # noqa
    if not dashboard:
        raise HTTPException(status_code=404)

    dashboard.is_archived = True
    session.commit()



from typing import Optional


class BaseNotifier(object):
    def __init__(self, _id: str):
        self._id = _id

    def __repr__(self):
        return f'<{self.__class__.__name__} object at {id(self)}>'

    def notify(self, message: Optional[str] = None, **kwargs):
        raise NotImplementedError



from aim.ext.cleanup import AutoClean
from aim.ext.transport.client import Client


class RemoteResourceAutoClean(AutoClean):
    def __init__(self, instance):
        super().__init__(instance)
        self.handler = None
        self.rpc_client: Client = None

    def _close(self):
        if self.handler is not None:
            assert self.rpc_client is not None
            self.rpc_client.release_resource(self.handler)



from aim import Repo
from aim.sdk.configs import get_aim_repo_name
from aim.sdk.types import QueryReportMode
from aim.storage.rockscontainer import RocksContainer

from performance_tests.utils import timing


@timing(10)
def open_containers_for_read(container_paths_list):
    sequences_subdir = f'{get_aim_repo_name()}/seqs/chunks/'
    for path in container_paths_list:
        container_path = f'{sequences_subdir}/{path}'
        container = RocksContainer(container_path, read_only=True)
        container.db


@timing()
def random_access_metric_values(repo, query, density):
    traces = repo.query_metrics(query=query, report_mode=QueryReportMode.DISABLED)
    for trace in traces.iter():
        values = trace.values
        values_length = len(values)
        step = len(values)//density

        accessed_values = []
        for i in range(0, values_length, step):
            accessed_values.append(trace.values[i])


@timing()
def iterative_access_metric_values(repo, query):
    traces = repo.query_metrics(query=query, report_mode=QueryReportMode.DISABLED)
    for trace in traces.iter():
        _ = trace.values.values_numpy()


def collect_sequence_containers():
    repo = Repo.default_repo()
    return [run.hash for run in repo.iter_runs()]



from abc import abstractmethod, ABC
from typing import Generic, TypeVar, Collection, Optional, List

T = TypeVar('T')

RunCollection = Collection['Run']
ExperimentCollection = Collection['Experiment']
TagCollection = Collection['Tag']
NoteCollection = Collection['Note']


class StructuredObject(ABC):
    @classmethod
    @abstractmethod
    def fields(cls):
        ...


class Searchable(ABC, Generic[T]):
    @classmethod
    @abstractmethod
    def find(cls, _id: str, **kwargs) -> Optional[T]:
        ...

    @classmethod
    @abstractmethod
    def all(cls, **kwargs) -> Collection[T]:
        ...

    @classmethod
    @abstractmethod
    def search(cls, term: str, **kwargs) -> Collection[T]:
        ...


class Run(StructuredObject, Searchable['Run']):
    @property
    @abstractmethod
    def hash(self) -> str:
        ...

    @property
    @abstractmethod
    def name(self) -> Optional[str]:
        ...

    @name.setter
    @abstractmethod
    def name(self, value: str):
        ...

    @property
    @abstractmethod
    def description(self) -> Optional[str]:
        ...

    @description.setter
    @abstractmethod
    def description(self, value: str):
        ...

    @property
    @abstractmethod
    def archived(self) -> bool:
        ...

    @archived.setter
    @abstractmethod
    def archived(self, value: bool):
        ...

    @property
    @abstractmethod
    def experiment(self) -> Optional['Experiment']:
        ...

    @experiment.setter
    @abstractmethod
    def experiment(self, value: str):
        ...

    @property
    @abstractmethod
    def tags(self) -> TagCollection:
        ...

    @abstractmethod
    def add_tag(self, value: str) -> 'Tag':
        ...

    @abstractmethod
    def remove_tag(self, tag_name: str) -> bool:
        ...

    @property
    @abstractmethod
    def info(self) -> 'RunInfo':
        ...


class Experiment(StructuredObject, Searchable['Experiment']):
    @property
    @abstractmethod
    def uuid(self) -> str:
        ...

    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @name.setter
    @abstractmethod
    def name(self, value: str):
        ...

    @property
    @abstractmethod
    def description(self) -> Optional[str]:
        ...

    @description.setter
    @abstractmethod
    def description(self, value: str):
        ...

    @property
    @abstractmethod
    def archived(self) -> bool:
        ...

    @archived.setter
    @abstractmethod
    def archived(self, value: bool):
        ...

    @property
    @abstractmethod
    def runs(self) -> RunCollection:
        ...


class Tag(StructuredObject, Searchable['Tag']):
    @property
    @abstractmethod
    def uuid(self) -> str:
        ...

    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @name.setter
    @abstractmethod
    def name(self, value: str):
        ...

    @property
    @abstractmethod
    def color(self) -> str:
        ...

    @color.setter
    @abstractmethod
    def color(self, value: str):
        ...

    @property
    @abstractmethod
    def description(self) -> str:
        ...

    @description.setter
    @abstractmethod
    def description(self, value: str):
        ...

    @property
    @abstractmethod
    def archived(self) -> bool:
        ...

    @archived.setter
    @abstractmethod
    def archived(self, value: bool):
        ...

    @property
    @abstractmethod
    def runs(self) -> RunCollection:
        ...


class Note(StructuredObject, Searchable['Note']):
    @property
    @abstractmethod
    def id(self) -> int:
        ...

    @property
    @abstractmethod
    def content(self) -> str:
        ...

    @content.setter
    @abstractmethod
    def content(self, value: str):
        ...

    @property
    @abstractmethod
    def run(self) -> int:
        ...


class RunInfo(StructuredObject, Generic[T]):
    @property
    @abstractmethod
    def last_notification_index(self) -> int:
        ...

    @last_notification_index.setter
    @abstractmethod
    def last_notification_index(self, value: int):
        ...


class ObjectFactory:
    @abstractmethod
    def runs(self) -> RunCollection:
        ...

    @abstractmethod
    def search_runs(self, term: str) -> RunCollection:
        ...

    @abstractmethod
    def find_run(self, _id: str) -> Run:
        ...

    @abstractmethod
    def find_runs(self, ids: List[str]) -> List[Run]:
        ...

    @abstractmethod
    def create_run(self, runhash: str) -> Run:
        ...

    @abstractmethod
    def delete_run(self, runhash: str) -> bool:
        ...

    @abstractmethod
    def experiments(self) -> ExperimentCollection:
        ...

    @abstractmethod
    def search_experiments(self, term: str) -> ExperimentCollection:
        ...

    @abstractmethod
    def find_experiment(self, _id: str) -> Experiment:
        ...

    @abstractmethod
    def create_experiment(self, name: str) -> Experiment:
        ...

    @abstractmethod
    def tags(self) -> TagCollection:
        ...

    @abstractmethod
    def search_tags(self, term: str) -> TagCollection:
        ...

    @abstractmethod
    def find_tag(self, _id: str) -> Tag:
        ...

    @abstractmethod
    def create_tag(self, name: str) -> Tag:
        ...

    @abstractmethod
    def delete_tag(self, name: str) -> bool:
        ...



import json
import os

import click
from tqdm import tqdm

from aim import Audio, Image, Run


def parse_tb_logs(tb_logs, repo_inst, flat=False, no_cache=False):
    """
    This function scans and collects records from TB log files.

    Creates and uses cache file "tb_logs_cache" in the repo dir
    to track previously processed files and values

    For more info please refer to our integration guides.
    """

    try:
        # This import statement takes long to complete
        import tensorflow as tf
        from tensorflow.python.summary.summary_iterator import summary_iterator
        from tensorboard.util import tensor_util
    except ImportError:
        click.echo(
            'Could not process TensorBoard logs - failed to import tensorflow module.', err=True
        )
        return

    supported_plugins = ('images', 'scalars')
    unsupported_plugin_noticed = False
    tb_logs_cache_path = os.path.join(repo_inst.path, 'tb_logs_cache')

    if no_cache and os.path.exists(tb_logs_cache_path):
        os.remove(tb_logs_cache_path)
    try:
        with open(tb_logs_cache_path) as FS:
            tb_logs_cache = json.load(FS)
    except Exception:
        tb_logs_cache = {}

    def get_parent(current_path, level=0):
        # level 0 is the direct parent directory
        if level <= 0:
            return os.path.dirname(current_path)
        elif current_path in ('', '.', '/'):
            return current_path
        return get_parent(os.path.dirname(current_path), level - 1)

    tb_logs = os.path.abspath(tb_logs)
    run_dir_candidates = set()
    for root, dirs, files in os.walk(tb_logs):
        for file in files:
            if not file.startswith('events.out.tfevents'):
                continue

            file_path = os.path.abspath(os.path.join(root, file))
            run_dir = get_parent(file_path)

            if not run_dir.startswith(tb_logs):
                # it's outside tb_logs
                continue

            run_dir_candidates.add(run_dir)

    def get_level(current_path):
        level = -1
        while current_path.startswith(tb_logs):
            current_path, _ = os.path.split(current_path)
            level += 1
        return level

    def create_ndarray(tensor):
        res = tensor_util.make_ndarray(tensor)
        if res.dtype == "object":
            return None
        else:
            return res

    run_dir_candidates = sorted(run_dir_candidates, key=get_level, reverse=True)
    run_dir_candidates_filtered = set()
    run_dir_ignored = set()
    groups = set()

    for run_dir in run_dir_candidates:
        if run_dir in run_dir_candidates_filtered:
            # already tagged as a run dir
            continue

        if run_dir in groups:
            # run dir which has other run dirs inside, so we skip it
            run_dir_ignored.add(run_dir)
            continue

        depth = get_level(run_dir)
        if depth >= 2:
            if flat:
                run_group_dir = get_parent(run_dir, 0)
                new_run_dir = run_dir
            else:
                run_group_dir = get_parent(run_dir, 1)
                new_run_dir = get_parent(run_dir, 0)
                if new_run_dir in groups:
                    new_run_dir = run_dir
            groups.add(run_group_dir)
        elif depth == 1:
            new_run_dir = run_dir
        else:
            continue
        run_dir_candidates_filtered.add(new_run_dir)

    if run_dir_ignored:
        click.echo('WARN: Found directory entries with unorganized even files!\n'
                   'Please read the preparation instructions to properly process these files.\n'
                   'Event files in the following directories will be ignored:', err=True)
        for c, r in enumerate(run_dir_ignored, start=1):
            click.echo(f'{c}: {r}', err=True)

    for path in tqdm(run_dir_candidates_filtered,
                     desc='Converting TensorBoard logs',
                     total=len(run_dir_candidates_filtered)):

        events = {}
        for root, dirs, files in os.walk(path):
            for file in files:
                if 'events.out.tfevents' not in file:
                    continue
                file_path = os.path.join(root, file)
                if file_path == os.path.join(path, file):
                    entry = None
                else:
                    entry = os.path.basename(os.path.dirname(file_path))
                events[file_path] = {
                    'context': {
                        'entry': entry
                    }
                }

        if path not in tb_logs_cache:
            tb_logs_cache[path] = {}

        run_cache = tb_logs_cache[path]
        if run_cache:
            run = Run(
                run_hash=run_cache['run_hash'],
                repo=repo_inst,
                system_tracking_interval=None,
                log_system_params=False,
                capture_terminal_logs=False,
            )
        else:
            run = Run(
                repo=repo_inst,
                system_tracking_interval=None,
                log_system_params=False,
                capture_terminal_logs=False,
            )
            run['tensorboard_logdir'] = path
            run_cache.update({
                'run_hash': run.hash,
                'events': {},
            })
        run_tb_events = run_cache['events']

        events_to_process = []
        for event in events:
            last_modified_at = os.path.getmtime(event)
            try:
                assert last_modified_at == run_tb_events[event]['last_modified_at']
            except (KeyError, AssertionError, RuntimeError):
                # Something has changed or hasn't been processed before
                events_to_process.append(event)
                try:
                    run_tb_events[event]['last_modified_at'] = last_modified_at
                except KeyError:
                    # Completely new event
                    run_tb_events[event] = {
                        'last_modified_at': last_modified_at,
                        'values': {},
                    }

        if not events_to_process:
            continue

        for event_file in tqdm(events_to_process, desc=f'Parsing logs in {path}', total=len(events_to_process)):
            run_tb_log = run_tb_events[event_file]
            event_context = events[event_file]['context']
            try:
                for event in summary_iterator(event_file):
                    timestamp = event.wall_time
                    step = event.step
                    fail_count = 0
                    _err_info = None

                    for value in event.summary.value:
                        tag = value.tag

                        plugin_name = value.metadata.plugin_data.plugin_name
                        value_id = f'{tag}_{plugin_name}'
                        if value_id in run_tb_log['values']:
                            if run_tb_log['values'][value_id]['timestamp'] >= timestamp:
                                # prevent previously tracked data from re-tracking upon file update
                                continue

                        if len(plugin_name) > 0 and plugin_name not in supported_plugins:
                            if not unsupported_plugin_noticed:
                                click.echo(
                                    'Found unsupported plugin type in the log file. '
                                    'Data for these wont be processed. '
                                    'Supported plugin types are: {}'.format(', '.join(supported_plugins)),
                                    err=True
                                )
                                unsupported_plugin_noticed = True
                            continue
                        track_val = None
                        try:
                            if value.HasField('tensor'):
                                # TODO: [MV] check the case when audios are passed via tensor
                                if plugin_name == 'images':
                                    tensor = value.tensor.string_val[2:]
                                    track_val = [
                                        Image(tf.image.decode_image(t).numpy()) for t in tensor
                                    ]
                                    if len(track_val) == 1:
                                        track_val = track_val[0]
                                elif plugin_name == "scalars" or plugin_name == "":
                                    track_val = create_ndarray(value.tensor)
                                else:
                                    track_val = value.tensor.float_val[0]
                            elif value.HasField('simple_value'):
                                track_val = value.simple_value
                            elif value.HasField('image'):
                                track_val = Image(tf.image.decode_image(value.image.encoded_image_string).numpy())
                            elif value.HasField('audio'):
                                tf_audio, sample_rate = tf.audio.decode_wav(value.audio.encoded_audio_string)
                                track_val = Audio(tf_audio.numpy(), rate=sample_rate)

                        except RuntimeError as exc:
                            # catch all the nasty failures
                            fail_count += 1
                            if not _err_info:
                                _err_info = str(exc)
                            continue

                        run_tb_log['values'][value_id] = {
                            'step': step,
                            'timestamp': timestamp
                        }
                        if track_val is not None:
                            run._tracker._track(track_val, timestamp, tag, step, context=event_context)
                    if fail_count:
                        click.echo(f'Failed to process {fail_count} entries. First exception: {_err_info}', err=True)

            except RuntimeError as exc:
                click.echo(f'Failed to read log file {event_file} - {exc}', err=True)

    # refresh cache
    with open(tb_logs_cache_path, 'w') as FS:
        json.dump(tb_logs_cache, FS)

    click.echo('TensorBoard logs conversion complete!')



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
# temporary workaround for M1 build
try:
    import grpc
except ImportError:
    grpc = None

import aim.ext.transport.remote_tracking_pb2 as remote__tracking__pb2


class RemoteTrackingServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.cleanup_client_resources = channel.unary_unary(
                '/RemoteTrackingService/cleanup_client_resources',
                request_serializer=remote__tracking__pb2.ClientResourceCleanupRequest.SerializeToString,
                response_deserializer=remote__tracking__pb2.ClientResourceCleanupResponse.FromString,
                )
        self.get_resource = channel.unary_unary(
                '/RemoteTrackingService/get_resource',
                request_serializer=remote__tracking__pb2.ResourceRequest.SerializeToString,
                response_deserializer=remote__tracking__pb2.ResourceResponse.FromString,
                )
        self.release_resource = channel.unary_unary(
                '/RemoteTrackingService/release_resource',
                request_serializer=remote__tracking__pb2.ReleaseResourceRequest.SerializeToString,
                response_deserializer=remote__tracking__pb2.ReleaseResourceResponse.FromString,
                )
        self.run_instruction = channel.stream_stream(
                '/RemoteTrackingService/run_instruction',
                request_serializer=remote__tracking__pb2.InstructionRequest.SerializeToString,
                response_deserializer=remote__tracking__pb2.InstructionResponse.FromString,
                )
        self.run_write_instructions = channel.stream_unary(
                '/RemoteTrackingService/run_write_instructions',
                request_serializer=remote__tracking__pb2.WriteInstructionsRequest.SerializeToString,
                response_deserializer=remote__tracking__pb2.WriteInstructionsResponse.FromString,
                )


class RemoteTrackingServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def cleanup_client_resources(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_resource(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def release_resource(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def run_instruction(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def run_write_instructions(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_RemoteTrackingServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'cleanup_client_resources': grpc.unary_unary_rpc_method_handler(
                    servicer.cleanup_client_resources,
                    request_deserializer=remote__tracking__pb2.ClientResourceCleanupRequest.FromString,
                    response_serializer=remote__tracking__pb2.ClientResourceCleanupResponse.SerializeToString,
            ),
            'get_resource': grpc.unary_unary_rpc_method_handler(
                    servicer.get_resource,
                    request_deserializer=remote__tracking__pb2.ResourceRequest.FromString,
                    response_serializer=remote__tracking__pb2.ResourceResponse.SerializeToString,
            ),
            'release_resource': grpc.unary_unary_rpc_method_handler(
                    servicer.release_resource,
                    request_deserializer=remote__tracking__pb2.ReleaseResourceRequest.FromString,
                    response_serializer=remote__tracking__pb2.ReleaseResourceResponse.SerializeToString,
            ),
            'run_instruction': grpc.stream_stream_rpc_method_handler(
                    servicer.run_instruction,
                    request_deserializer=remote__tracking__pb2.InstructionRequest.FromString,
                    response_serializer=remote__tracking__pb2.InstructionResponse.SerializeToString,
            ),
            'run_write_instructions': grpc.stream_unary_rpc_method_handler(
                    servicer.run_write_instructions,
                    request_deserializer=remote__tracking__pb2.WriteInstructionsRequest.FromString,
                    response_serializer=remote__tracking__pb2.WriteInstructionsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'RemoteTrackingService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class RemoteTrackingService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def cleanup_client_resources(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RemoteTrackingService/cleanup_client_resources',
            remote__tracking__pb2.ClientResourceCleanupRequest.SerializeToString,
            remote__tracking__pb2.ClientResourceCleanupResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_resource(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RemoteTrackingService/get_resource',
            remote__tracking__pb2.ResourceRequest.SerializeToString,
            remote__tracking__pb2.ResourceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def release_resource(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RemoteTrackingService/release_resource',
            remote__tracking__pb2.ReleaseResourceRequest.SerializeToString,
            remote__tracking__pb2.ReleaseResourceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def run_instruction(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/RemoteTrackingService/run_instruction',
            remote__tracking__pb2.InstructionRequest.SerializeToString,
            remote__tracking__pb2.InstructionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def run_write_instructions(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/RemoteTrackingService/run_write_instructions',
            remote__tracking__pb2.WriteInstructionsRequest.SerializeToString,
            remote__tracking__pb2.WriteInstructionsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import logging
from abc import abstractmethod
from typing import Iterator
from typing import TYPE_CHECKING
from tqdm import tqdm

from aim.sdk.sequence import Sequence
from aim.sdk.types import QueryReportMode
from aim.sdk.query_utils import RunView, SequenceView
from aim.storage.query import RestrictedPythonQuery


if TYPE_CHECKING:
    from aim.sdk.run import Run
    from aim.sdk.repo import Repo
    from pandas import DataFrame

logger = logging.getLogger(__name__)


class SequenceCollection:
    """Abstract interface for collection of tracked series/sequences.

    Typically represents sequences of a same run or sequences matching given query expression.
    """

    # TODO [AT]: move to a separate mixin class
    def dataframe(
        self,
        only_last: bool = False,
        include_run=True,
        include_name=True,
        include_context=True,
        include_props=True,
        include_params=True,
    ) -> 'DataFrame':
        # TODO [GA]: Separate runs and sequences dataframes collection
        dfs = []
        if self._item == 'run':
            dfs = [
                run.run.dataframe(include_props=include_props,
                                  include_params=include_params)
                for run in self.iter_runs()
            ]
        elif self._item == 'sequence':
            dfs = [
                metric.dataframe(include_run=include_run,
                                 include_name=include_name,
                                 include_context=include_context,
                                 only_last=only_last)
                for metric in self
            ]
        if not dfs:
            return None
        import pandas as pd
        return pd.concat(dfs)

    def __iter__(self) -> Iterator[Sequence]:
        return self.iter()

    @abstractmethod
    def iter(self) -> Iterator[Sequence]:
        """Get Sequence iterator for collection's sequences.

        Yields:
            Next sequence object based on implementation.
        """
        ...

    @abstractmethod
    def iter_runs(self) -> Iterator['SequenceCollection']:
        """Get SequenceCollection iterator for collection's runs.

        Yields:
            Next run's SequenceCollection based on implementation.
        """
        ...


class SingleRunSequenceCollection(SequenceCollection):
    """Implementation of SequenceCollection interface for a single Run.

    Method `iter()` returns Sequence iterator which yields Sequence matching query from run's sequences.
    Method `iter_runs()` raises StopIteration, since the collection is bound to a single Run.

    Args:
         run (:obj:`Run`): Run object for which sequences are queried.
         seq_cls (:obj:`type`): The collection's sequence class. Sequences not matching to seq_cls.allowed_dtypes
            will be skipped. `Sequence` by default, meaning all sequences will match.
         query (:obj:`str`, optional): Query expression. If specified, method `iter()` will return iterator for
            sequences matching the query. If not, method `iter()` will return iterator for run's all sequences.
    """
    def __init__(
        self,
        run: 'Run',
        seq_cls=Sequence,
        query: str = '',
        runs_proxy_cache: dict = None,
        timezone_offset: int = 0,
    ):
        self.run: 'Run' = run
        self.seq_cls = seq_cls
        self._item = 'sequence'
        self.query = RestrictedPythonQuery(query)
        self.runs_proxy_cache = runs_proxy_cache
        self._timezone_offset = timezone_offset

    def iter_runs(self) -> Iterator['SequenceCollection']:
        """"""
        logger.warning('Run is already bound to the Collection')
        raise StopIteration

    def iter(
        self
    ) -> Iterator[Sequence]:
        """"""
        allowed_dtypes = self.seq_cls.allowed_dtypes()
        seq_var = self.seq_cls.sequence_name()
        for seq_name, ctx, run in self.run.iter_sequence_info_by_type(allowed_dtypes):
            run_view = RunView(run, self.runs_proxy_cache, self._timezone_offset)
            seq_view = SequenceView(seq_name, ctx.to_dict(), run_view)
            match = self.query.check(**{'run': run_view, seq_var: seq_view})
            if not match:
                continue
            yield self.seq_cls(seq_name, ctx, run)


class QuerySequenceCollection(SequenceCollection):
    """Implementation of SequenceCollection interface for repository's sequences matching given query.

    Method `iter()` returns Sequence iterator, which yields Sequence matching query from currently
    iterated run's sequences. Once there are no sequences left in current run, repository's next run is considered.
    Method `iter_runs()` returns SequenceCollection iterator for repository's runs.

    Args:
         repo (:obj:`Repo`): Aim repository object.
         seq_cls (:obj:`type`): The collection's sequence class. Sequences not matching to seq_cls.allowed_dtypes
            will be skipped. `Sequence` by default, meaning all sequences will match.
         query (:obj:`str`, optional): Query expression. If specified, method `iter()` will skip sequences not matching
            the query. If not, method `iter()` will return iterator for all sequences in repository
            (that's a lot of sequences!).
    """

    def __init__(
        self,
        repo: 'Repo',
        seq_cls=Sequence,
        query: str = '',
        report_mode: QueryReportMode = QueryReportMode.PROGRESS_BAR,
        timezone_offset: int = 0,
    ):
        self.repo: 'Repo' = repo
        self.seq_cls = seq_cls
        self._item = 'sequence'
        self.query = query
        self.report_mode = report_mode
        self.runs_proxy_cache = dict()
        self._timezone_offset = timezone_offset

    def iter_runs(self) -> Iterator['SequenceCollection']:
        """"""
        if self.repo.structured_db:
            runs_iterator = self.repo.iter_runs_from_cache()
        else:
            runs_iterator = self.repo.iter_runs()
        runs_counter = 1
        total_runs = self.repo.total_runs_count()

        if self.report_mode == QueryReportMode.PROGRESS_BAR:
            progress_bar = tqdm(total=total_runs)

        for run in runs_iterator:
            seq_collection = SingleRunSequenceCollection(run, self.seq_cls, self.query,
                                                         runs_proxy_cache=self.runs_proxy_cache,
                                                         timezone_offset=self._timezone_offset)
            if self.report_mode == QueryReportMode.PROGRESS_TUPLE:
                yield seq_collection, (runs_counter, total_runs)
            else:
                if self.report_mode == QueryReportMode.PROGRESS_BAR:
                    progress_bar.update(1)
                yield seq_collection
            runs_counter += 1

    def iter(self) -> Iterator[Sequence]:
        """"""
        if self.report_mode == QueryReportMode.PROGRESS_TUPLE:
            for run_seq, _ in self.iter_runs():
                yield from run_seq
        else:
            for run_seq in self.iter_runs():
                yield from run_seq


class QueryRunSequenceCollection(SequenceCollection):
    """Implementation of SequenceCollection interface for repository's runs matching given query.

    Method `iter()` returns Sequence iterator which yields Sequence for current run's all sequences.
    Method `iter_runs()` returns SequenceCollection iterator from repository's runs matching given query.

    Args:
         repo (:obj:`Repo`): Aim repository object.
         seq_cls (:obj:`type`): The collection's sequence class. Sequences not matching to seq_cls.allowed_dtypes
            will be skipped. `Sequence` by default, meaning all sequences will match.
         query (:obj:`str`, optional): Query expression. If specified, method `iter_runs()` will skip runs not matching
            the query. If not, method `iter_run()` will return SequenceCollection iterator for all runs in repository.
    """

    def __init__(
        self,
        repo: 'Repo',
        seq_cls=Sequence,
        query: str = '',
        paginated: bool = False,
        offset: str = None,
        report_mode: QueryReportMode = QueryReportMode.PROGRESS_BAR,
        timezone_offset: int = 0,
    ):
        self.repo: 'Repo' = repo
        self.seq_cls = seq_cls
        self.query = query
        self._item = 'run'
        self.paginated = paginated
        self.offset = offset
        self.query = RestrictedPythonQuery(query)
        self.report_mode = report_mode
        self._timezone_offset = timezone_offset

    def iter(self) -> Iterator[Sequence]:
        """"""
        if self.report_mode == QueryReportMode.PROGRESS_TUPLE:
            for run_seq, _ in self.iter_runs():
                yield from run_seq
        else:
            for run_seq in self.iter_runs():
                yield from run_seq

    def iter_runs(self) -> Iterator['SequenceCollection']:
        """"""
        if self.repo.structured_db:
            runs_iterator = self.repo.iter_runs_from_cache(offset=self.offset)
        else:
            runs_iterator = self.repo.iter_runs()
        runs_counter = 1
        total_runs = self.repo.total_runs_count()
        if self.report_mode == QueryReportMode.PROGRESS_BAR:
            progress_bar = tqdm(total=total_runs)
        for run in runs_iterator:
            run_view = RunView(run, timezone_offset=self._timezone_offset)
            match = self.query.check(run=run_view)
            seq_collection = SingleRunSequenceCollection(run, self.seq_cls) if match else None
            if self.report_mode == QueryReportMode.PROGRESS_TUPLE:
                yield seq_collection, (runs_counter, total_runs)
            else:
                if self.report_mode == QueryReportMode.PROGRESS_BAR:
                    progress_bar.update(1)
                if match:
                    yield seq_collection
            runs_counter += 1



from typing import Tuple

from aim.sdk.sequence import Sequence
from aim.sdk.objects.distribution import Distribution


class Distributions(Sequence):
    """Class representing series of Distribution objects."""

    @classmethod
    def allowed_dtypes(cls) -> Tuple[str, ...]:
        return (Distribution.get_typename(),)

    @classmethod
    def sequence_name(cls) -> str:
        return 'distributions'

    def first_step(self):
        """Get sequence tracked first step.

        Required to implement ranged and sliced data fetching.
        """
        return self._meta_tree['first_step']

    def last_step(self):
        """Get sequence tracked last step.

        Required to implement ranged and sliced data fetching.
        """
        return self._meta_tree['last_step']



import functools
from typing import Callable, Optional, Sequence, Union

import optuna
from optuna._experimental import experimental_class, experimental_func
from optuna._imports import try_import
from optuna.study.study import ObjectiveFuncType


with try_import() as _imports:
    from aim.sdk.run import Run
    from aim.ext.resource.configs import DEFAULT_SYSTEM_TRACKING_INT


@experimental_class("2.9.0")
class AimCallback:
    """Callback to track Optuna trials with Aim.

    This callback enables tracking of Optuna study in Aim.
    The study is tracked as a single experiment run,
    where all suggested hyperparameters and optimized metrics
    are logged and plotted as a function of optimizer steps.

    .. note::
        Users who want to run multiple Optuna studies within the same process
        should call ``close()`` method between subsequent calls to
        ``study.optimize()``. Calling ``close()`` method is not necessary
        if you are running one Optuna study per process.

    .. note::
        To ensure correct trial order in Aim, this callback
        should only be used with ``study.optimize(n_jobs=1)``.


    Args:
        metric_name:
            Name assigned to optimized metric. In case of multi-objective optimization,
            list of names can be passed. Those names will be assigned
            to metrics in the order returned by objective function.
            If single name is provided, or this argument is left to default value,
            it will be broadcasted to each objective with a number suffix in order
            returned by objective function e.g. two objectives and default metric name
            will be logged as ``value_0`` and ``value_1``. The number of metrics must be
            the same as the number of values objective function returns.
        as_multirun:
            Creates new runs for each trial and sets the metrics as run parametrs.
            Useful for exploring parameters in Aim UI
            (for more: https://aimstack.readthedocs.io/en/latest/ui/pages/explorers.html#params-explorer).
            If is false then all of the stats are tracked in a single run as Aim metrics.

    """

    def __init__(
        self,
        metric_name: Union[str, Sequence[str]] = "value",
        as_multirun: bool = False,
        repo: Optional[str] = None,
        experiment_name: Optional[str] = None,
        system_tracking_interval: Optional[int] = DEFAULT_SYSTEM_TRACKING_INT,
        log_system_params: bool = True,
    ) -> None:

        _imports.check()

        if not isinstance(metric_name, Sequence):
            raise TypeError(
                f"Expected metric_name to be string or sequence of strings, got {type(metric_name)}."
            )

        self._metric_name = metric_name
        self._as_multirun = as_multirun
        self._repo_path = repo
        self._experiment_name = experiment_name
        self._system_tracking_interval = system_tracking_interval
        self._log_system_params = log_system_params
        self._run = None
        self._run_hash = None

        if not self._as_multirun:
            self.setup()

    def __call__(self, study: optuna.study.Study, trial: optuna.trial.FrozenTrial) -> None:

        if isinstance(self._metric_name, str):
            if len(trial.values) > 1:
                # Broadcast default name for multi-objective optimization.
                names = [f"{self._metric_name}_{i}" for i in range(len(trial.values))]
            else:
                names = [self._metric_name]
        else:
            if len(self._metric_name) != len(trial.values):
                raise ValueError(
                    "Running multi-objective optimization "
                    f"with {len(trial.values)} objective values, but {len(self._metric_name)} names specified. "
                    "Match objective values and names, or use default broadcasting."
                )
            else:
                names = self._metric_name

        metrics = {name: value for name, value in zip(names, trial.values)}

        if self._as_multirun:
            metrics["trial_number"] = trial.number

        attributes = {"direction": [d.name for d in study.directions]}

        step = trial.number if self._run else None

        if not self._run:
            self.setup()
            self._run.name = f"trial-{trial.number}"

        for key, value in attributes.items():
            self._run.set(('hparams', key), value, strict=False)

        self._run.set("study_name", study.study_name)

        if self._as_multirun:
            for key, value in trial.params.items():
                self._run.set(('hparams', key), value, strict=False)

            for key, value in metrics.items():
                self._run.set(key, value, strict=False)

            self.close()
        else:
            for key, value in trial.params.items():
                self._run.track(value, name=key, step=step)

            for key, value in metrics.items():
                self._run.track(value, name=key, step=step)

    @property
    def experiment(self) -> Run:
        if self._run is not None:
            return self._run

    def setup(self, args=None):
        if not self._run:
            if self._run_hash:
                self._run = Run(
                    self._run_hash,
                    repo=self._repo_path,
                    system_tracking_interval=self._system_tracking_interval,
                )
            else:
                self._run = Run(
                    repo=self._repo_path,
                    experiment=self._experiment_name,
                    system_tracking_interval=self._system_tracking_interval,
                    log_system_params=self._log_system_params,
                )
                self._run_hash = self._run.hash

        if args:
            for key, value in args.items():
                self._run.set(key, value, strict=False)

    def close(self) -> None:
        if self._run:
            self._run.close()
            self._run = None
            self._run_hash = None

    @experimental_func("3.0.0")
    def track_in_aim(self) -> Callable:
        """Decorator for using Aim for tracking inside the objective function.

        All the metrics from inside the objective function will be logged into the same run
        which stores the parameters for a given trial.

        Returns:
            ObjectiveFuncType: Objective function with Aim tracking enabled.
        """

        def decorator(func: ObjectiveFuncType) -> ObjectiveFuncType:
            @functools.wraps(func)
            def wrapper(trial: optuna.trial.Trial) -> Union[float, Sequence[float]]:
                if not self._run:
                    self.setup()
                    self._run.name = f"trial-{trial.number}"
                return func(trial)
            return wrapper
        return decorator



import datetime
import pytz

from typing import Any, Union
from typing import TYPE_CHECKING

from aim.storage.context import Context
from aim.storage.proxy import AimObjectProxy
from aim.storage.structured.entities import StructuredObject
from aim.storage.treeview import TreeView
from aim.storage.types import AimObject, AimObjectKey, AimObjectPath, SafeNone
from aim.storage.structured.sql_engine.entities import ModelMappedRun

if TYPE_CHECKING:
    from aim.sdk.run import Run


class RunView:
    def __init__(self, run: 'Run', runs_proxy_cache: dict = None, timezone_offset: int = 0):
        self.db = run.repo.structured_db
        self.hash = run.hash
        self.structured_run_cls: type(StructuredObject) = ModelMappedRun
        self.meta_run_tree: TreeView = run.meta_run_tree
        self.meta_run_attrs_tree: TreeView = run.meta_run_attrs_tree
        self.run = run
        self.proxy_cache = None
        self._timezone_offset = timezone_offset
        if runs_proxy_cache is not None:
            if runs_proxy_cache.get(run.hash) is None:
                runs_proxy_cache[run.hash] = {}
            self.proxy_cache = runs_proxy_cache[run.hash]

    def __getattr__(self, item):
        if item == 'metrics':
            return MetricsView(self.meta_run_tree, self.proxy_cache)
        if item in ['finalized_at', 'end_time']:
            end_time = self.meta_run_tree['end_time']
            if item == 'finalized_at':
                if not end_time:
                    return None
                else:
                    return datetime.datetime.fromtimestamp(end_time, tz=pytz.utc).replace(tzinfo=None)\
                        - datetime.timedelta(minutes=self._timezone_offset)
            else:
                return end_time
        if item == 'created_at':
            return getattr(self.db.caches['runs_cache'][self.hash], item)\
                - datetime.timedelta(minutes=self._timezone_offset)
        if item in ('active', 'duration'):
            return getattr(self.run, item)
        elif item in self.structured_run_cls.fields():
            if self.db:
                return getattr(self.db.caches['runs_cache'][self.hash], item)
            else:
                return getattr(self.run, item)
        else:
            return self[item]

    def __getitem__(self, key):
        def safe_collect():
            res = None
            if self.proxy_cache is not None:
                res = self.proxy_cache.get(key)
            if not res:
                try:
                    res = self.meta_run_attrs_tree.collect(key)
                except Exception:
                    res = SafeNone()
                if self.proxy_cache is not None:
                    self.proxy_cache[key] = res
            return res

        return AimObjectProxy(safe_collect, view=self.meta_run_attrs_tree.subtree(key),
                              cache=self.proxy_cache)

    def get(
        self,
        key,
        default: Any = None
    ) -> AimObject:
        try:
            return self.__getitem__(key)
        except KeyError:
            return default


class MetricsView:
    def __init__(self, meta_run_tree, proxy_cache):
        self.meta_run_tree = meta_run_tree
        self.proxy_cache = proxy_cache

    def __getitem__(self, item):
        def safe_collect():
            res = None
            if self.proxy_cache is not None:
                res = self.proxy_cache.get(key)
            if not res:
                try:
                    res = self.meta_run_attrs_tree.collect(key)
                except Exception:
                    res = SafeNone()
                if self.proxy_cache is not None:
                    self.proxy_cache[key] = res
            return res

        if isinstance(item, str):
            metric_name = item
            context_idx = Context({}).idx
        elif isinstance(item, tuple):
            if len(item) > 2:
                return SafeNone()
            metric_name, context = item
            if not isinstance(metric_name, str):
                return SafeNone()
            if isinstance(context, int):
                context_idx = context
            elif isinstance(context, dict):
                context_idx = Context(context).idx
            else:
                return SafeNone()
        else:
            return SafeNone()

        key = ('traces', context_idx, metric_name)
        return AimObjectProxy(safe_collect, view=self.meta_run_tree.subtree(key), cache=self.proxy_cache)

    def __getattr__(self, item):
        raise NotImplementedError


class ContextView:
    def __init__(self, context: dict):
        self.context = context

    def __getitem__(self, key):
        return self.context[key]

    def get(
            self,
            key,
            default: Any = None
    ) -> AimObject:
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def view(self, path: Union[AimObjectKey, AimObjectPath]):
        if isinstance(path, (int, str)):
            path = (path,)

        return ContextView(self.context[path[0]])


class SequenceView:
    def __init__(self, name: str, context: dict, run_view: RunView):
        self.name = name
        self.run = run_view
        self._context = context
        self._sequence_meta_tree = None

    @property
    def context(self):
        return AimObjectProxy(lambda: self._context, view=ContextView(self._context))

    def __getattr__(self, item):
        def safe_collect():
            try:
                return self._sequence_meta_tree.collect(item)
            except Exception:
                return SafeNone()

        if not self._sequence_meta_tree:
            self._sequence_meta_tree = self.run.meta_run_tree.subtree(('traces',
                                                                       Context(self._context).idx,
                                                                       self.name))

        return AimObjectProxy(safe_collect, view=self._sequence_meta_tree.subtree(item))



"""empty message

Revision ID: 5ae8371b7481
Revises: 11672b13f92c
Create Date: 2021-06-10 14:13:40.552387

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '5ae8371b7481'
down_revision = '11672b13f92c'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('dashboards',
    sa.Column('uuid', sa.Text(), nullable=False),
    sa.Column('name', sa.Text(), nullable=True),
    sa.Column('description', sa.Text(), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('is_archived', sa.Boolean(), nullable=True),
    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('explore_states',
    sa.Column('uuid', sa.Text(), nullable=False),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('is_archived', sa.Boolean(), nullable=True),
    sa.Column('chart_focused_step', sa.Integer(), nullable=True),
    sa.Column('chart_focused_metric_run_hash', sa.Text(), nullable=True),
    sa.Column('chart_focused_metric_metric_name', sa.Text(), nullable=True),
    sa.Column('chart_focused_metric_trace_context', sa.Text(), nullable=True),
    sa.Column('chart_focused_circle_active', sa.Boolean(), nullable=True),
    sa.Column('chart_focused_circle_run_hash', sa.Text(), nullable=True),
    sa.Column('chart_focused_circle_metric_name', sa.Text(), nullable=True),
    sa.Column('chart_focused_circle_step', sa.Integer(), nullable=True),
    sa.Column('chart_focused_circle_trace_context', sa.Text(), nullable=True),
    sa.Column('chart_focused_circle_param', sa.Text(), nullable=True),
    sa.Column('chart_focused_circle_content_type', sa.Text(), nullable=True),
    sa.Column('chart_settings_zoom_mode', sa.Boolean(), nullable=True),
    sa.Column('chart_settings_single_zoom_mode', sa.Boolean(), nullable=True),
    sa.Column('chart_settings_zoom_history', sa.Text(), nullable=True),
    sa.Column('chart_settings_highlight_mode', sa.Text(), nullable=True),
    sa.Column('chart_settings_persistent_display_outliers', sa.Boolean(), nullable=True),
    sa.Column('chart_settings_persistent_zoom', sa.Text(), nullable=True),
    sa.Column('chart_settings_persistent_interpolate', sa.Boolean(), nullable=True),
    sa.Column('chart_settings_persistent_indicator', sa.Boolean(), nullable=True),
    sa.Column('chart_settings_persistent_x_alignment', sa.Text(), nullable=True),
    sa.Column('chart_settings_persistent_x_scale', sa.Integer(), nullable=True),
    sa.Column('chart_settings_persistent_y_scale', sa.Integer(), nullable=True),
    sa.Column('chart_settings_persistent_points_count', sa.Integer(), nullable=True),
    sa.Column('chart_settings_persistent_smoothing_algorithm', sa.Text(), nullable=True),
    sa.Column('chart_settings_persistent_smooth_factor', sa.Float(), nullable=True),
    sa.Column('chart_settings_persistent_aggregated', sa.Boolean(), nullable=True),
    sa.Column('chart_hidden_metrics', sa.Text(), nullable=True),
    sa.Column('chart_tooltip_options_display', sa.Boolean(), nullable=True),
    sa.Column('chart_tooltip_options_fields', sa.Text(), nullable=True),
    sa.Column('search_query', sa.Text(), nullable=True),
    sa.Column('search_v', sa.Integer(), nullable=True),
    sa.Column('search_input_value', sa.Text(), nullable=True),
    sa.Column('search_input_select_input', sa.Text(), nullable=True),
    sa.Column('search_input_select_condition_input', sa.Text(), nullable=True),
    sa.Column('context_filter_group_by_color', sa.Text(), nullable=True),
    sa.Column('context_filter_group_by_style', sa.Text(), nullable=True),
    sa.Column('context_filter_group_by_chart', sa.Text(), nullable=True),
    sa.Column('context_filter_group_against_color', sa.Boolean(), nullable=True),
    sa.Column('context_filter_group_against_style', sa.Boolean(), nullable=True),
    sa.Column('context_filter_group_against_chart', sa.Boolean(), nullable=True),
    sa.Column('context_filter_aggregated_area', sa.Text(), nullable=True),
    sa.Column('context_filter_aggregated_line', sa.Text(), nullable=True),
    sa.Column('context_filter_seed_color', sa.Integer(), nullable=True),
    sa.Column('context_filter_seed_style', sa.Integer(), nullable=True),
    sa.Column('context_filter_persist_color', sa.Boolean(), nullable=True),
    sa.Column('context_filter_persist_style', sa.Boolean(), nullable=True),
    sa.Column('color_palette', sa.Integer(), nullable=True),
    sa.Column('sort_fields', sa.Text(), nullable=True),
    sa.Column('table_row_height_mode', sa.Text(), nullable=True),
    sa.Column('table_columns_order_left', sa.Text(), nullable=True),
    sa.Column('table_columns_order_middle', sa.Text(), nullable=True),
    sa.Column('table_columns_order_right', sa.Text(), nullable=True),
    sa.Column('table_columns_widths', sa.Text(), nullable=True),
    sa.Column('table_excluded_fields', sa.Text(), nullable=True),
    sa.Column('screen_view_mode', sa.Text(), nullable=True),
    sa.Column('screen_panel_flex', sa.Float(), nullable=True),
    sa.Column('dashboard_id', sa.Text(), nullable=True),
    sa.ForeignKeyConstraint(['dashboard_id'], ['dashboards.uuid'], ),
    sa.PrimaryKeyConstraint('uuid')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('explore_states')
    op.drop_table('dashboards')
    # ### end Alembic commands ###



from aim.storage.types import AimObject, AimObjectKey, AimObjectPath

from typing import TYPE_CHECKING, Any, Iterator, Tuple, Union, List

if TYPE_CHECKING:
    from aim.storage.arrayview import ArrayView


class TreeView:

    def preload(
        self
    ):
        ...

    def finalize(
        self,
        index: 'TreeView'
    ):
        ...

    def subtree(self, path: Union[AimObjectKey, AimObjectPath]) -> 'TreeView':
        # Default to:
        return self.view(path, resolve=False)

    def view(
        self,
        path: Union[AimObjectKey, AimObjectPath],
        resolve: bool = False
    ):
        ...

    def make_array(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ):
        ...

    def collect(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        strict: bool = True,
        resolve_objects: bool = False
    ) -> AimObject:
        ...

    def __getitem__(
        self,
        path: Union[AimObjectKey, AimObjectPath]
    ) -> AimObject:
        return self.collect(path)

    def get(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        default: Any = None
    ) -> AimObject:
        try:
            return self[path]
        except KeyError:
            return default

    def __delitem__(
        self,
        path: Union[AimObjectKey, AimObjectPath]
    ):
        ...

    def set(
        self,
        path: Union[AimObjectKey, AimObjectPath],
        value: AimObject,
        strict: bool = True
    ):
        self.__setitem__(path, value)

    def __setitem__(
        self,
        path: Union[AimObjectKey, AimObjectPath],
        value: AimObject
    ):
        self.set(path, value, strict=True)

    # TODO [MV]: revisit this approach after cython bindings refactoring
    def keys_eager(
            self,
            path: Union[AimObjectKey, AimObjectPath] = (),
    ) -> List[Union[AimObjectPath, AimObjectKey]]:
        ...

    def keys(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        level: int = None
    ) -> Iterator[Union[AimObjectPath, AimObjectKey]]:
        ...

    def items_eager(
            self,
            path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> List[Tuple[
        AimObjectKey,
        AimObject
    ]]:
        ...

    def items(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> Iterator[Tuple[
        AimObjectKey,
        AimObject
    ]]:
        ...

    def array(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        dtype: Any = None
    ) -> 'ArrayView':
        ...

    def first_key(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> AimObjectKey:
        ...

    def last_key(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> AimObjectKey:
        ...



# Aim UI
AIM_UI_DEFAULT_PORT = 43800
AIM_UI_DEFAULT_HOST = '127.0.0.1'
AIM_TF_LOGS_PATH_KEY = '__AIM_TF_LOGS_PATH__'
AIM_ENV_MODE_KEY = '__AIM_ENV_MODE__'
AIM_LOG_LEVEL_KEY = '__AIM_LOG_LEVEL__'
AIM_UI_MOUNTED_REPO_PATH = '__AIM_UI_MOUNT_REPO_PATH__'
AIM_UI_TELEMETRY_KEY = 'AIM_UI_TELEMETRY_ENABLED'
AIM_UI_BASE_PATH = '__AIM_UI_BASE_PATH__'
AIM_PROXY_URL = '__AIM_PROXY_URL__'
AIM_PROFILER_KEY = '__AIM_PROFILER_ENABLED__'
AIM_PROGRESS_REPORT_INTERVAL = 0.5
AIM_PROJECT_SETTINGS_FILE = '.project_settings'



import os
import psutil
import json
from typing import List

from aim.ext.resource.utils import round10e5

from py3nvml import py3nvml as nvml


class StatDict(object):
    # Available aggregation functions
    AGG_MODE_AVG = 'average'
    AGG_MODE_MIN = 'min'
    AGG_MODE_MAX = 'max'
    AGG_MODE_DIFF = 'diff'
    AGG_DEFAULT = AGG_MODE_AVG

    @classmethod
    def aggregate(cls, items: List, mode: str):
        """
        Aggregates array of numbers by a given 'mode'
        """
        if mode == cls.AGG_MODE_MAX:
            return max(items)
        elif mode == cls.AGG_MODE_MIN:
            return min(items)
        elif mode == cls.AGG_MODE_AVG:
            return round10e5(sum(items) / len(items))
        elif mode == cls.AGG_MODE_DIFF:
            return round10e5(max(items) - min(items))
        else:
            raise ValueError('unknown aggregation mode: \'{}\''.format(mode))

    @classmethod
    def aggregate_items(cls,
                        items: 'List[StatDict]',
                        agg_mode: str = AGG_DEFAULT,
                        ):
        """
        Aggregates array of `StatDict` items by a given `mode`
        """
        aggregated_stat = cls()

        # Return empty item if items array is empty
        if not items or len(items) == 0:
            return aggregated_stat

        gpu_stats = []
        for s in items:
            # Collect system stats
            for k in s.system.keys():
                aggregated_stat.system.setdefault(k, [])
                aggregated_stat.system[k].append(s.system[k])

            # Collect GPU device stats
            for stat_item_gpu_idx in range(len(s.gpus)):
                stat_item_gpu_stat = s.gpus[stat_item_gpu_idx]
                if len(gpu_stats) == stat_item_gpu_idx:
                    gpu_stats.append({})
                for gpu_stat_key in stat_item_gpu_stat.keys():
                    gpu_stat = stat_item_gpu_stat[gpu_stat_key]
                    gpu_stats[stat_item_gpu_idx].setdefault(gpu_stat_key, [])
                    gpu_stats[stat_item_gpu_idx][gpu_stat_key].append(gpu_stat)

        # Aggregate system stats
        for k in aggregated_stat.system.keys():
            aggregated_stat.system[k] = cls.aggregate(aggregated_stat.system[k],
                                                      agg_mode)

        # Aggregate GPU device stats
        for g in range(len(gpu_stats)):
            for k in gpu_stats[g].keys():
                gpu_stats[g][k] = cls.aggregate(gpu_stats[g][k], agg_mode)
        aggregated_stat.gpu = gpu_stats

        return aggregated_stat

    def __init__(self, system: dict = None, gpus: List[dict] = None):
        self.system = system or {}
        self.gpus = gpus or []

    def __str__(self):
        return json.dumps(self.to_dict())

    def to_dict(self):
        """
        Returns system and GPU device statistics
        """
        return {
            'system': self.system,
            'gpus': self.gpus,
        }


class Stat(object):
    def __init__(self, process):
        # Set process
        self._process = process

        # Get statistics
        system, gpus = self.get_stats()
        self._stat = StatDict(system, gpus)

    @property
    def process(self):
        return self._process

    @property
    def stat_item(self):
        return self._stat

    @property
    def system(self):
        return self._stat.system

    @property
    def gpus(self):
        return self._stat.gpus

    def get_stats(self):
        """
        Get system statistics and assign to `self`
        """
        memory_usage = psutil.virtual_memory()
        disk_usage = psutil.disk_usage('/')
        # net = psutil.net_io_counters()
        system = {
            # CPU utilization percent(can be over 100%)
            'cpu': psutil.cpu_percent(0.0) if os.getenv('AIM_GLOBAL_CPU_PERCENTAGE') == 'true' else round10e5(self._process.cpu_percent(0.0)),

            # Whole system memory usage
            # 'memory_used': round10e5(memory_usage.used / 1024 / 1024),
            'memory_percent': round10e5(memory_usage.used * 100 / memory_usage.total),

            # Get the portion of memory occupied by a process
            # 'p_memory_rss': round10e5(self._process.memory_info().rss
            #                           / 1024 / 1024),
            'p_memory_percent': round10e5(self._process.memory_percent()),

            # Disk usage
            # 'disk_used': round10e5(disk_usage.used / 1024 / 1024),
            'disk_percent': round10e5(disk_usage.percent),
        }

        # Collect GPU statistics
        gpus = []
        try:
            nvml.nvmlInit()
            gpu_device_count = nvml.nvmlDeviceGetCount()
            for i in range(gpu_device_count):
                gpu_info = dict()
                handle = nvml.nvmlDeviceGetHandleByIndex(i)
                try:
                    util = nvml.nvmlDeviceGetUtilizationRates(handle)
                    # GPU utilization percent
                    gpu_info["gpu"] = round10e5(util.gpu)
                except nvml.NVMLError_NotSupported:
                    pass
                try:
                    # Get device memory
                    memory = nvml.nvmlDeviceGetMemoryInfo(handle)
                    # Device memory usage
                    # 'memory_used': round10e5(memory.used / 1024 / 1024),
                    gpu_info["gpu_memory_percent"] = (
                        round10e5(memory.used * 100 / memory.total),
                    )
                except nvml.NVMLError_NotSupported:
                    pass
                try:
                    # Get device temperature
                    nvml_tmp = nvml.NVML_TEMPERATURE_GPU
                    temp = nvml.nvmlDeviceGetTemperature(handle, nvml_tmp)
                    # Device temperature
                    gpu_info["gpu_temp"] = round10e5(temp)
                except nvml.NVMLError_NotSupported:
                    pass
                try:
                    # Compute power usage in watts and percent
                    power_watts = nvml.nvmlDeviceGetPowerUsage(handle) / 1000
                    power_cap = nvml.nvmlDeviceGetEnforcedPowerLimit(handle)
                    power_cap_watts = power_cap / 1000
                    power_watts / power_cap_watts * 100
                    # Power usage in watts and percent
                    gpu_info["gpu_power_watts"]: round10e5(power_watts)
                    # gpu_info["power_percent"] = round10e5(power_usage)
                except nvml.NVMLError_NotSupported:
                    pass
                gpus.append(gpu_info)
            nvml.nvmlShutdown()
        except (nvml.NVMLError_LibraryNotFound, nvml.NVMLError_NotSupported):
            pass

        return system, gpus



"""empty message

Revision ID: 517a45b2e62c
Revises: 73a3d004c227
Create Date: 2021-08-31 20:58:39.490140

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '517a45b2e62c'
down_revision = '73a3d004c227'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('dashboards',
    sa.Column('uuid', sa.Text(), nullable=False),
    sa.Column('name', sa.Text(), nullable=True),
    sa.Column('description', sa.Text(), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('is_archived', sa.Boolean(), nullable=True),
    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('explore_states',
    sa.Column('uuid', sa.Text(), nullable=False),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('is_archived', sa.Boolean(), nullable=True),
    sa.Column('type', sa.Text(), nullable=False),
    sa.Column('state', sa.Text(), nullable=True),
    sa.Column('dashboard_id', sa.Text(), nullable=True),
    sa.ForeignKeyConstraint(['dashboard_id'], ['dashboards.uuid'], ),
    sa.PrimaryKeyConstraint('uuid')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('explore_states')
    op.drop_table('dashboards')
    # ### end Alembic commands ###



import logging
import os

from contextlib import contextmanager
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from aim.web.configs import AIM_LOG_LEVEL_KEY
from aim.web.utils import get_db_url

engine = create_engine(
    get_db_url(),
    echo=(logging.INFO >= int(os.environ.get(AIM_LOG_LEVEL_KEY, logging.WARNING))),
    connect_args={"check_same_thread": False}
)
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)
Base = declarative_base()


def get_session():
    session = SessionLocal()
    try:
        yield session
    finally:
        session.close()


@contextmanager
def get_contexted_session():
    session = SessionLocal()
    try:
        yield session
    finally:
        session.close()



import os
import time

from typing import Optional
from concurrent import futures

from aim.ext.cleanup import AutoClean

from aim.ext.transport.config import AIM_RT_MAX_MESSAGE_SIZE, AIM_RT_DEFAULT_MAX_MESSAGE_SIZE
from aim.ext.transport.handlers import get_tree, get_structured_run, get_repo, get_lock, get_run_heartbeat
from aim.ext.transport.heartbeat import RPCHeartbeatWatcher
from aim.ext.transport.worker import RemoteWorker, LocalWorker
from aim.ext.transport.health import HealthServicer, health_pb2_grpc
from aim.ext.transport.router import RemoteRouterServicer, router_pb2_grpc
from aim.ext.transport.remote_tracking import RemoteTrackingServicer, tracking_pb2_grpc, ResourceTypeRegistry


def _wait_forever():
    while True:
        time.sleep(24 * 60 * 60)  # sleep for a day


class RPCServerAutoClean(AutoClean['RPCServer']):
    PRIORITY = 150

    def __init__(self, instance: 'RPCServer'):
        super().__init__(instance)
        self.server = instance.server

    def _close(self):
        self.server.stop(grace=None)


class RPCServer:
    def __init__(self):
        import grpc  # temporary workaround for M1 build

        self._resources: Optional[RPCServerAutoClean] = None
        interceptors = tuple()
        if os.getenv('AIM_ENABLE_AUTH'):
            from .server_auth_interceptor import ServerAuthInterceptor
            interceptors = (ServerAuthInterceptor(),)
        self._server = grpc.server(futures.ThreadPoolExecutor(), interceptors=interceptors, options=self._get_grpc_server_options())
        self.resource = RPCServerAutoClean(self)

    def start(self, host, port, ssl_keyfile, ssl_certfile):
        import grpc  # temporary workaround for M1 build

        if ssl_keyfile and ssl_certfile:
            with open(ssl_keyfile, 'rb') as f:
                private_key = f.read()
            with open(ssl_certfile, 'rb') as f:
                certificate_chain = f.read()
            server_credentials = grpc.ssl_server_credentials([(private_key, certificate_chain,)])
            self._server.add_secure_port(f'{host}:{port}', server_credentials)
        else:
            self._server.add_insecure_port(f'{host}:{port}')

        self._server.start()

    @staticmethod
    def _get_grpc_server_options():
        msg_max_size = int(os.getenv(AIM_RT_MAX_MESSAGE_SIZE, AIM_RT_DEFAULT_MAX_MESSAGE_SIZE))
        options = [
            ('grpc.max_send_message_length', msg_max_size),
            ('grpc.max_receive_message_length', msg_max_size)
        ]
        return options

    @property
    def server(self):
        return self._server


def prepare_resource_registry():
    registry = ResourceTypeRegistry()
    registry.register('TreeView', get_tree)
    registry.register('StructuredRun', get_structured_run)
    registry.register('Repo', get_repo)
    registry.register('Lock', get_lock)
    registry.register('RunHeartbeat', get_run_heartbeat)
    return registry


def run_router(host, port, workers=1, ssl_keyfile=None, ssl_certfile=None):
    # start workers
    worker_pool = []

    if workers == 1:
        worker = LocalWorker(lambda: None, host=host, port=port, index=0,
                             ssl_keyfile=ssl_keyfile, ssl_certfile=ssl_certfile)
        worker_pool.append(worker)
    else:
        for i in range(1, workers + 1):
            worker_port = port + i
            worker = RemoteWorker(run_worker, host=host, port=worker_port, index=i,
                                  ssl_keyfile=ssl_keyfile, ssl_certfile=ssl_certfile)
            worker_pool.append(worker)
            worker.start()

    # start routing RPC server
    server = RPCServer()
    router_pb2_grpc.add_RemoteRouterServiceServicer_to_server(RemoteRouterServicer(worker_pool), server.server)
    health_pb2_grpc.add_HealthServicer_to_server(HealthServicer(), server.server)

    if workers == 1:
        # add worker servicers to router as well
        registry = prepare_resource_registry()
        tracking_pb2_grpc.add_RemoteTrackingServiceServicer_to_server(RemoteTrackingServicer(registry), server.server)

    server.start(host, port, ssl_keyfile, ssl_certfile)

    # start client heartbeat service
    heartbeat_watcher = RPCHeartbeatWatcher(
        RemoteRouterServicer.client_heartbeat_pool,
        worker_pool,
    )
    heartbeat_watcher.start()

    _wait_forever()


def run_worker(host, port, ssl_keyfile=None, ssl_certfile=None):
    # register resource handlers
    registry = prepare_resource_registry()

    # start tracking RPC server
    server = RPCServer()
    tracking_pb2_grpc.add_RemoteTrackingServiceServicer_to_server(RemoteTrackingServicer(registry), server.server)
    server.start(host, port, ssl_keyfile, ssl_certfile)

    _wait_forever()



from datetime import timedelta
from collections import Counter
from fastapi import Request, HTTPException, Depends, Header
from aim.web.api.utils import APIRouter  # wrapper for fastapi.APIRouter
from typing import Optional

from aim.web.api.projects.project import Project
from aim.web.api.utils import object_factory
from aim.web.api.experiments.pydantic_models import (
    ExperimentGetOut,
    ExperimentUpdateOut,
    ExperimentGetRunsOut,
    ExperimentListOut,
    ExperimentCreateIn,
    ExperimentUpdateIn,
    ExperimentActivityApiOut,
)
from aim.web.api.runs.pydantic_models import NoteIn

experiment_router = APIRouter()

# static msgs
NOTE_NOT_FOUND = 'Note with id {id} is not found in this experiment.'


@experiment_router.get('/', response_model=ExperimentListOut)
async def get_experiments_list_api(factory=Depends(object_factory)):
    return [
        {'id': exp.uuid,
         'name': exp.name,
         'description': exp.description,
         'run_count': len(exp.runs),
         'archived': exp.archived,
         'creation_time': exp.creation_time
         } for exp in factory.experiments()]


@experiment_router.get('/search/', response_model=ExperimentListOut)
async def search_experiments_by_name_api(request: Request, factory=Depends(object_factory)):
    params = request.query_params
    search_term = params.get('q') or ''
    search_term.strip()

    return [{'id': exp.uuid, 'name': exp.name, 'run_count': len(exp.runs), 'archived': exp.archived}
            for exp in factory.search_experiments(search_term)]


@experiment_router.post('/', response_model=ExperimentUpdateOut)
async def create_experiment_api(exp_in: ExperimentCreateIn, factory=Depends(object_factory)):
    with factory:
        try:
            exp = factory.create_experiment(exp_in.name.strip())
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e))

    return {
        'id': exp.uuid,
        'status': 'OK'
    }


@experiment_router.get('/{exp_id}/', response_model=ExperimentGetOut)
async def get_experiment_api(exp_id: str, factory=Depends(object_factory)):
    exp = factory.find_experiment(exp_id)
    if not exp:
        raise HTTPException(status_code=404)

    response = {
        'id': exp.uuid,
        'name': exp.name,
        'description': exp.description,
        'archived': exp.archived,
        'run_count': len(exp.runs),
        'creation_time': exp.creation_time
    }
    return response


@experiment_router.put('/{exp_id}/', response_model=ExperimentUpdateOut)
async def update_experiment_properties_api(exp_id: str, exp_in: ExperimentUpdateIn, factory=Depends(object_factory)):
    with factory:
        exp = factory.find_experiment(exp_id)
        if not exp:
            raise HTTPException(status_code=404)

        if exp_in.name:
            exp.name = exp_in.name.strip()
        if exp_in.description is not None:
            exp.description = exp_in.description
        if exp_in.archived is not None:
            if exp_in.archived and len(exp.runs) > 0:
                raise HTTPException(status_code=400, detail=(
                    f'Cannot archive experiment \'{exp_id}\'. '
                    'Experiment has associated runs.'
                ))
            exp.archived = exp_in.archived

    return {
        'id': exp.uuid,
        'status': 'OK'
    }


@experiment_router.get('/{exp_id}/runs/', response_model=ExperimentGetRunsOut)
async def get_experiment_runs_api(exp_id: str,
                                  limit: Optional[int] = None,
                                  offset: Optional[str] = None,
                                  factory=Depends(object_factory)):
    project = Project()

    exp = factory.find_experiment(exp_id)
    if not exp:
        raise HTTPException(status_code=404)

    from aim.sdk.run import Run

    cache_name = 'exp_runs'
    project.repo.run_props_cache_hint = cache_name
    project.repo.structured_db.invalidate_cache(cache_name)
    project.repo.structured_db.init_cache(cache_name, exp.get_runs, lambda run_: run_.hash)
    exp_runs = []

    run_hashes = [run.hash for run in exp.runs]
    offset_idx = 0
    if offset:
        try:
            offset_idx = run_hashes.index(offset) + 1
        except ValueError:
            pass
    if limit:
        run_hashes = run_hashes[offset_idx: offset_idx + limit]

    for run_hash in run_hashes:
        run = Run(run_hash, repo=project.repo, read_only=True)
        exp_runs.append({
            'run_id': run.hash,
            'name': run.name,
            'creation_time': run.creation_time,
            'end_time': run.end_time,
            'archived': run.archived
        })

    project.repo.structured_db.invalidate_cache(cache_name)
    project.repo.run_props_cache_hint = None

    response = {
        'id': exp.uuid,
        'runs': exp_runs
    }
    return response

# Note APIs


@experiment_router.get('/{exp_id}/note/')
async def list_note_api(exp_id, factory=Depends(object_factory)):
    with factory:
        experiment = factory.find_experiment(exp_id)
        if not experiment:
            raise HTTPException(status_code=404)

        notes = experiment.notes

    return notes


@experiment_router.post('/{exp_id}/note/', status_code=201)
async def create_note_api(exp_id, note_in: NoteIn, factory=Depends(object_factory)):
    with factory:
        experiment = factory.find_experiment(exp_id)
        if not experiment:
            raise HTTPException(status_code=404)

        note_content = note_in.content.strip()
        note = experiment.add_note(note_content)

    return {
        'id': note.id,
        'created_at': note.created_at,
    }


@experiment_router.get('/{exp_id}/note/{_id}')
async def get_note_api(exp_id, _id: int, factory=Depends(object_factory)):
    with factory:
        experiment = factory.find_experiment(exp_id)
        if not experiment:
            raise HTTPException(status_code=404)

        note = experiment.find_note(_id=_id)
        if not note:
            raise HTTPException(status_code=404, detail=NOTE_NOT_FOUND.format(id=_id))

    return {
        'id': note.id,
        'content': note.content,
        'updated_at': note.updated_at,
    }


@experiment_router.put('/{exp_id}/note/{_id}')
async def update_note_api(exp_id, _id: int, note_in: NoteIn, factory=Depends(object_factory)):
    with factory:
        experiment = factory.find_experiment(exp_id)
        if not experiment:
            raise HTTPException(status_code=404)

        note = experiment.find_note(_id=_id)
        if not note:
            raise HTTPException(status_code=404, detail=NOTE_NOT_FOUND.format(id=_id))

        content = note_in.content.strip()
        updated_note = experiment.update_note(_id=_id, content=content)

    return {
        'id': updated_note.id,
        'content': updated_note.content,
        'updated_at': updated_note.updated_at,
    }


@experiment_router.delete('/{exp_id}/note/{_id}')
async def delete_note_api(exp_id, _id: int, factory=Depends(object_factory)):
    with factory:
        experiment = factory.find_experiment(exp_id)
        if not experiment:
            raise HTTPException(status_code=404)

        note = experiment.find_note(_id=_id)
        if not note:
            raise HTTPException(status_code=404, detail=NOTE_NOT_FOUND.format(id=_id))

        experiment.remove_note(_id)

    return {
        'status': 'OK'
    }


@experiment_router.get('/{exp_id}/activity/', response_model=ExperimentActivityApiOut)
async def experiment_runs_activity_api(exp_id,
                                       x_timezone_offset: int = Header(default=0),
                                       factory=Depends(object_factory)):
    project = Project()

    if not project.exists():
        raise HTTPException(status_code=404)

    with factory:
        experiment = factory.find_experiment(exp_id)
        if not experiment:
            raise HTTPException(status_code=404)

        active_run_hashes = project.repo.list_active_runs()

        num_runs = 0
        num_archived_runs = 0
        num_active_runs = 0
        activity_counter = Counter()

        for run in experiment.runs:
            creation_time = run.created_at - timedelta(minutes=x_timezone_offset)
            activity_counter[creation_time.strftime('%Y-%m-%dT%H:00:00')] += 1
            num_runs += 1
            if run.archived:
                num_archived_runs += 1
            if run.hash in active_run_hashes:
                num_active_runs += 1

        return {
            'num_runs': num_runs,
            'num_archived_runs': num_archived_runs,
            'num_active_runs': num_active_runs,
            'activity_map': dict(activity_counter),
        }



# distutils: language = c++
# cython: language_level=3
# distutils: language_level = 3

from aim.storage.hashing cimport c_hash

cpdef Py_hash_t det_hash(double v):
    """Custom hashing function for numeric values.
    Heavily based on CPython implementation with few details.

    Similar CPython, we expect `hash(x) == hash(y)` if `x == y`
    even if one of them is `int` and other one is `float`.
    This is achieved be mapping any floating-point number to a integer such that
    if the number encodes integer value it maps to itself with one exception:
      *  CPython uses a reserved hash value of `-1` for errors. So they map both
         integer `-1` and float `-1.0` to hash value `-2`.
    This is not a big issue when using them for hashmaps / hashsets but it
    introduces to frequent collisions.
    For example, even when CPython's PYTHONHASHSEED is used, it is easy to
    verify that: `hash((-1, -3)) == hash((-2, -3))`. When mapping `-1` to `-2`
    is avoided, hash seeds can really make a huge difference.
    """
    return c_hash.le_Py_HashDouble(v)



import os

from collections import defaultdict
from weakref import WeakValueDictionary

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session

from aim.storage.migrations.utils import upgrade_database
from aim.storage.structured.sql_engine.factory import ModelMappedFactory as ObjectFactory
from aim.storage.types import SafeNone
from aim.web.configs import AIM_LOG_LEVEL_KEY


class ObjectCache:
    def __init__(self, data_fetch_func, key_func):
        self._data = defaultdict(SafeNone)
        self._cached = False
        self.data_fetch_func = data_fetch_func
        self.key_func = key_func

    def fill_cache(self):
        for obj in self.data_fetch_func():
            self._data[self.key_func(obj)] = obj

    def keys(self) -> list:
        if not self._cached:
            self.fill_cache()
            self._cached = True
        return list(self._data.keys())

    def __setitem__(self, key, value):
        assert self._cached
        self._data[key] = value

    def __getitem__(self, key):
        if not self._cached:
            self.fill_cache()
            self._cached = True
        return self._data[key]


class DB(ObjectFactory):
    _DIALECT = 'sqlite'
    _DB_NAME = 'run_metadata.sqlite'
    _pool = WeakValueDictionary()

    _caches = dict()

    # TODO: [AT] implement readonly if needed
    def __init__(self, path: str, readonly: bool = False):
        import logging
        super().__init__()
        self.path = path
        self.db_url = self.get_db_url(path)
        self.readonly = readonly
        self.engine = create_engine(self.db_url,
                                    echo=(logging.INFO >= int(os.environ.get(AIM_LOG_LEVEL_KEY, logging.WARNING))))
        self.session_cls = scoped_session(sessionmaker(autocommit=False, autoflush=False, bind=self.engine))
        self._upgraded = None

    @classmethod
    def from_path(cls, path: str, readonly: bool = False):
        db = cls._pool.get(path)
        if not db:
            db = DB(path, readonly)
            cls._pool[path] = db
        return db

    @staticmethod
    def get_default_url():
        return DB.get_db_url('.aim')

    @staticmethod
    def get_db_url(path: str) -> str:
        if os.path.exists(path):
            db_url = f'{DB._DIALECT}:///{path}/{DB._DB_NAME}'
            return db_url
        else:
            raise RuntimeError(f'Cannot find database {path}. Please init first.')

    @property
    def caches(self):
        return self._caches

    def get_session(self, autocommit=True):
        session = self.session_cls()
        session.autocommit = autocommit
        return session

    def run_upgrades(self):
        if self._upgraded:
            return
        upgrade_database(self.db_url)
        self._upgraded = True

    def init_cache(self, cache_name, callback, key_func):
        if cache_name in self._caches:
            return
        self._caches[cache_name] = ObjectCache(callback, key_func)

    def invalidate_all_caches(self):
        self._caches.clear()

    def invalidate_cache(self, cache_name):
        if self._caches.get(cache_name):
            del self._caches[cache_name]



from typing import TYPE_CHECKING
from aim.storage.treeutils import encode_tree
from aim.ext.transport.message_utils import pack_args

from aim.ext.transport.remote_resource import RemoteResourceAutoClean

if TYPE_CHECKING:
    from aim.ext.transport.client import Client


class RemoteRepoAutoClean(RemoteResourceAutoClean):
    PRIORITY = 60


class RemoteRepoProxy:
    def __init__(self, client: 'Client'):
        self._rpc_client = client

        self.init_args = pack_args(encode_tree(()))
        self.resource_type = 'Repo'

        handler = self._rpc_client.get_resource_handler(self, self.resource_type, args=self.init_args)

        self._resources = RemoteRepoAutoClean(self)
        self._resources.rpc_client = client
        self._resources.handler = handler
        self._handler = handler

    def list_all_runs(self):
        return self._rpc_client.run_instruction(-1, self._handler, 'list_all_runs', [])

    def list_active_runs(self):
        return self._rpc_client.run_instruction(-1, self._handler, 'list_active_runs', [])



import os
import shutil

from functools import lru_cache
from pathlib import Path
from typing import Union

here = os.path.abspath(os.path.dirname(__file__))


@lru_cache()
def get_working_directory(base_dir: Union[str, Path]) -> Path:
    if not isinstance(base_dir, Path):
        base_dir = Path(base_dir)
    work_dir = base_dir / 'ext' / 'notifications'
    work_dir.mkdir(parents=True, exist_ok=True)
    return work_dir


def get_config_path(base_dir: Union[str, Path]) -> Path:
    return get_working_directory(base_dir) / 'config.json'


def get_default_config_path() -> Path:
    return Path(here) / 'config_default.json'


def get_empty_config_path() -> Path:
    return Path(here) / 'config_empty.json'


def has_watcher_config(base_dir: Union[str, Path]) -> bool:
    config_path = get_config_path(base_dir)
    return config_path.is_file()


def set_default_config(base_dir: Union[str, Path]):
    config = get_config_path(base_dir)
    default_config = get_default_config_path()
    shutil.copy(default_config, config)



from logging import getLogger
from typing import Optional
from aim.sdk.run import Run

from aim.ext.resource.configs import DEFAULT_SYSTEM_TRACKING_INT

try:
    from fastai.learner import Callback
    from fastcore.basics import store_attr, detuplify, ignore_exceptions
    from fastai.callback.hook import total_params
except ImportError:
    raise RuntimeError(
        'This contrib module requires fastai to be installed. '
        'Please install it with command: \n pip install fastai'
    )

logger = getLogger(__name__)


class AimCallback(Callback):
    def __init__(self,
                 repo: Optional[str] = None,
                 experiment_name: Optional[str] = None,
                 system_tracking_interval: Optional[int] = DEFAULT_SYSTEM_TRACKING_INT,
                 log_system_params: bool = True,
                 ):
        store_attr()
        self._run = None
        self._run_hash = None

    @property
    def experiment(self) -> Run:
        if not self._run:
            self.setup()
        return self._run

    def setup(self, args=None):
        if not self._run:
            if self._run_hash:
                self._run = Run(
                    self._run_hash,
                    repo=self.repo,
                    system_tracking_interval=self.system_tracking_interval,
                    log_system_params=self.log_system_params,
                )
            else:
                self._run = Run(
                    repo=self.repo,
                    experiment=self.experiment_name,
                    system_tracking_interval=self.system_tracking_interval,
                    log_system_params=self.log_system_params,
                )
                self._run_hash = self._run.hash

        # Log config parameters
        if args:
            try:
                for key in args:
                    self._run.set(key, args[key], strict=False)
            except Exception as e:
                logger.warning(f'Aim could not log config parameters -> {e}')

    def before_fit(self):
        if not self._run:
            configs_log = self.gather_args()
            formatted_config = AimCallback.format_config(configs_log)
            self.setup(formatted_config)

    def after_batch(self):
        context = {'subset': 'train'} if self.training else {'subset': 'val'}

        self._run.track(self.loss.item(), name='train_loss', step=self.train_iter, context=context)
        for i, h in enumerate(self.opt.hypers):
            for k, v in h.items():
                self._run.track(v, f'{k}_{i}', step=self.train_iter, context=context)

    def before_epoch(self):
        for metric in self.metrics:
            metric.reset()

    def after_epoch(self):
        for name, value in zip(self.recorder.metric_names, self.recorder.log):
            if name not in ['train_loss', 'epoch', 'time'] and value is not None:
                self._run.track(value, name=name)

    def __del__(self):
        if self._run and self._run.active:
            self._run.close()

    def gather_args(self):
        "Gather config parameters accessible to the learner"
        # args stored by 'store_attr'
        cb_args = {f'{cb}': getattr(cb, '__stored_args__', True) for cb in self.cbs if cb != self}
        args = {'Learner': self.learn, **cb_args}
        # input dim
        try:
            n_inp = self.dls.train.n_inp
            args['n_inp'] = n_inp
            xb = self.dls.valid.one_batch()[:n_inp]
            args.update({f'input {n+1} dim {i+1}': d for n in range(n_inp)
                        for i, d in enumerate(list(detuplify(xb[n]).shape))})
        except Exception:
            logger.warning('Failed to gather input dimensions')
        # other args
        with ignore_exceptions():
            args['batch_size'] = self.dls.bs
            args['batch_per_epoch'] = len(self.dls.train)
            args['model_parameters'] = total_params(self.model)[0]
            args['device'] = self.dls.device.type
            args['frozen'] = bool(self.opt.frozen_idx)
            args['frozen_idx'] = self.opt.frozen_idx
            args['dataset', 'tfms'] = f'{self.dls.dataset.tfms}'
            args['dls', 'after_item'] = f'{self.dls.after_item}'
            args['dls', 'before_batch'] = f'{self.dls.before_batch}'
            args['dls', 'after_batch'] = f'{self.dls.after_batch}'
        return args

    @classmethod
    def format_config(cls, config):
        "Format config parameters for logging"
        for key, value in config.items():
            if isinstance(value, dict):
                config[key] = AimCallback.format_config(value)
            else:
                config[key] = AimCallback.format_config_value(value)
        return config

    @classmethod
    def format_config_value(cls, value):
        if isinstance(value, list):
            return [AimCallback.format_config_value(item) for item in value]
        elif hasattr(value, '__stored_args__'):
            return {**AimCallback.format_config(value.__stored_args__), '_name': value}
        return value



import pytest
import numpy as np

from tests.base import TestBase

from aim.sdk import Repo, Run
from aim.sdk.types import QueryReportMode


class TestRunMetricNumpyTypes(TestBase):
    @pytest.mark.gh_1206
    def test_numpy_scalar_types_track(self):
        """covers https://github.com/aimhubio/aim/issues/1206"""

        run = self.create_run(system_tracking_interval=None)
        run.track(np.array([1.0]), name='single_item_array')
        run.track(np.array([[[1.0]]]), name='single_item_3d_array')
        run.track(np.float64(1.0), name='numpy_float64')
        run.track(np.float32(1.0), name='numpy_float32')
        run.close()

        repo = Repo.default_repo()
        q = self.isolated_query_patch()
        metric_names = {metric.name for metric in repo.query_metrics(query=q, report_mode=QueryReportMode.DISABLED)}
        expected_metric_names = {'single_item_array', 'single_item_3d_array', 'numpy_float64', 'numpy_float32'}
        self.assertSetEqual(expected_metric_names, metric_names)

    def test_reject_non_scalar_arrays_track(self):
        run = Run(system_tracking_interval=None)
        with self.assertRaises(ValueError):
            run.track(np.array([1.0, 2.0]), name='fail')



# distutils: language = c++
# cython: language_level = 3

from aim.storage.treeview cimport TreeView

from aim.storage cimport encoding as E
from aim.storage.encoding.encoding cimport decode
from aim.storage.container cimport Container

cdef class ContainerTreeView(TreeView):
    cdef public Container container



def track_params_dists(model, run):
    from aim import Distribution
    data_hist = get_model_layers(model, 'data')

    for name, params in data_hist.items():
        if 'weight' in params:
            run.track(
                Distribution(params['weight']),
                name=name,
                context={
                    'type': 'data',
                    'params': 'weights',
                }
            )
        if 'bias' in params:
            run.track(
                Distribution(params['bias']),
                name=name,
                context={
                    'type': 'data',
                    'params': 'biases',
                }
            )


def track_gradients_dists(model, run):
    from aim import Distribution
    grad_hist = get_model_layers(model, 'grad')

    for name, params in grad_hist.items():
        if 'weight' in params:
            run.track(
                Distribution(params['weight']),
                name=name,
                context={
                    'type': 'gradients',
                    'params': 'weights',
                }
            )
        if 'bias' in params:
            run.track(
                Distribution(params['bias']),
                name=name,
                context={
                    'type': 'gradients',
                    'params': 'biases',
                }
            )


def get_model_layers(model, dt, parent_name=None):
    layers = {}
    for name, m in model.named_children():
        layer_name = '{}__{}'.format(parent_name, name) \
            if parent_name \
            else name
        layer_name += '.{}'.format(type(m).__name__)

        if len(list(m.named_children())):
            layers.update(get_model_layers(m, dt, layer_name))
        else:
            layers[layer_name] = {}
            weight = None
            if hasattr(m, 'weight') and m.weight is not None:
                weight = getattr(m.weight, dt, None)
                if weight is not None:
                    layers[layer_name]['weight'] = get_pt_tensor(weight).numpy()

            bias = None
            if hasattr(m, 'bias') and m.bias is not None:
                bias = getattr(m.bias, dt, None)
                if bias is not None:
                    layers[layer_name]['bias'] = get_pt_tensor(bias).numpy()

    return layers


# Move tensor from GPU to CPU
def get_pt_tensor(t):
    return t.cpu() if hasattr(t, 'is_cuda') and t.is_cuda else t



# distutils: language = c++
# cython: language_level = 3

"""Encodings for primitive values for AimObject-compatible data.

This includes None, booleans, integer and floating-point numbers, strings,
byte-array blobs and control flags for array and dict-like objects.

The encodings are implemented `(key, value)` design in mind, so using them to
stream and store in chunks is easy to manage.
"""

from aim.storage.encoding.encoding_native cimport PATH_SENTINEL_CODE
from aim.storage.encoding.encoding_native cimport (
    encode_int64,
    encode_double,
    encode_utf_8_str,
    encode_int64_big_endian,
)
from aim.storage.encoding.encoding_native cimport (
    decode_int64,
    decode_double,
    decode_utf_8_str,
)
from aim.storage.encoding.encoding_native cimport decode_path  # noqa F401
from aim.storage.utils import ArrayFlagType, ObjectFlagType, CustomObjectFlagType
from aim.storage.utils import ArrayFlag, ObjectFlag
from aim.storage.container import ContainerValue
from aim.storage.types import AimObjectKey, AimObjectPath
from aim.storage.types import BLOB

from typing import Union, Any

# The choice of `PATH_SENTINEL` is not arbitrary.
# This sentinel 1) consists of a single byte
#               2) does not appear in any utf-8 encoded string
#               3) comes right before max byte 0xff which make a room for
#                  many tricks
_PATH_SENTINEL = b'\xfe'
assert _PATH_SENTINEL[0] == PATH_SENTINEL_CODE


cpdef object encode(object value):
    """Automatically detect and encode the value into a buffer.
    This function is to encode only primitive objects.

    Type id of the value is encoded with the content as well in order to enable
    automatic decoding.
    The first byte of the buffer encodes type of the encoded value and the rest
    are for encoded content.
    """
    cdef int type_id
    cdef bytes encoding

    if isinstance(value, BLOB):
        return value.transform(encode)

    if value is None:
        # No need to encode a content for None-values
        type_id = FLAGS._NONE
        encoding = b''
    elif isinstance(value, bool):
        # Booleans are encoded in a single byte
        # 0 and 1 for False and True respectively
        type_id = FLAGS._BOOL
        encoding = b'\x01' if value else b'\x00'
    elif isinstance(value, int):
        # We encode integers in signed 64-bit
        type_id = FLAGS._INT
        encoding = encode_int64(value)
    elif isinstance(value, float):
        # We encode floats in double-precision
        type_id = FLAGS._FLOAT
        encoding = encode_double(value)
    elif isinstance(value, str):
        # Strings are stored in utf-8
        type_id = FLAGS._STRING
        encoding = encode_utf_8_str(value)
    elif isinstance(value, bytes):
        # Byte arrays / BLOBS can be used to encode arbitrary binary content
        type_id = FLAGS._BYTES
        encoding = value
    elif isinstance(value, ArrayFlagType):
        # Array flag is used to denote non-finised object, expecting
        # encodings for its elements
        type_id = FLAGS._ARRAY
        encoding = b''
    elif isinstance(value, ObjectFlagType):
        # Object flag is used to denote non-finised object, expecting
        # encodings for its elements
        type_id = FLAGS._OBJECT
        encoding = b''
    elif isinstance(value, CustomObjectFlagType):
        type_id = FLAGS._CUSTOM_OBJECT
        encoding = encode_utf_8_str(value.aim_name)
    else:
        raise NotImplementedError

    # Finally, we prepend the content `encoding` with a single byte which
    # encodes the type of the value
    cdef char* prefix = [type_id]
    return <bytes>prefix[:1] + encoding


cpdef object decode(object buffer):
    """Automatically detect and decode the value from a buffer."""

    if isinstance(buffer, BLOB):
        return buffer.transform(decode)

    # First, we extract type_id and the content buffer
    cdef int type_id = buffer[0]
    cdef bytes content_buffer = buffer[1:]
    # TODO pass the whole buffer to decode_* functions with offset

    # We check for type_id and decode the content for the corresponding type
    if type_id == FLAGS._NONE:
        return None
    elif type_id == FLAGS._BOOL:
        return content_buffer[0] != 0
    elif type_id == FLAGS._INT:
        return decode_int64(content_buffer)
    elif type_id == FLAGS._FLOAT:
        return decode_double(content_buffer)
    elif type_id == FLAGS._STRING:
        return decode_utf_8_str(content_buffer)
    elif type_id == FLAGS._BYTES:
        return content_buffer
    elif type_id == FLAGS._ARRAY:
        return ArrayFlag
    elif type_id == FLAGS._OBJECT:
        return ObjectFlag
    elif type_id == FLAGS._CUSTOM_OBJECT:
        return CustomObjectFlagType(decode_utf_8_str(content_buffer))
    else:
        return None


cpdef bytes encode_key(object key):
    """Encode the value as key in the path.

    This is different from what the general `encode()` will result.
    Here we omit type_id for strings to avoid redundancies, and add an
    additional path sentinel for numeric (integer) ones.
    Integers are encoded in big-endian to make the encoded keys to preserve
    the relative order when comparing as byte array buffers in lexicographic
    order.
    For example, `1000 > 249` but having them encoded in little-endian may have
    lead their encoded versions not to maintain the relative order:
    `b'\\xe8\\x03\\x00\\x00\\x00\\x00\\x00\\x00' < b'\\xf9\\x00\\x00\\x00\\x00\\x00\\x00\\x00'`
    However, in big-endiang encoding:
    `b'\\x00\\x00\\x00\\x00\\x00\\x00\\x03\\xe8' > b'\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\xf9'`
    """
    prefix = <object>bytes([PATH_SENTINEL_CODE])
    if isinstance(key, str):
        encoded = encode_utf_8_str(<str>key)
        return encoded
    elif isinstance(key, int):
        return prefix + encode_int64_big_endian(key)
    else:
        raise ValueError(f'Value {key} of type {type(key)} is not supported')


cpdef bytes encode_path(object path):
    """Encode the path into a buffer.
    The encoding of path is designed to effectively use prefix-trees that may
    be used by storage engines.

    The path is encoded as concatenation of encoded keys, separated with
    so-called `PATH_SENTINEL`.
    Keys can be either strings or integers:
      *  Strings keys are encoded in utf-8. Any non-empty string is allowed as
         a path key, containing of any characters. There is no risk of
         ambiguous encodings because the `PATH_SENTINEL` is chosen such it can
         never be occurred in correctly encoded utf-8 string. Such key may
         include even `PATH_SENTINEL` as it is encoded using other bytes.
      *  Integers are encoded in signed 64-bit integers, and a `PATH_SENTINEL`
         is prepended.
    If the key encoding starts with the `PATH_SENTINEL` it can only encode an
    integer, as 1) no utf-8 encoded string may start with `PATH_SENTINEL`
                2) the empty strings are not allowed
    """
    cdef tuple path_tuple
    if isinstance(path, (int, str)):
        path_tuple = (path,)
    elif isinstance(path, list):
        path_tuple = tuple(path)
    elif isinstance(path, tuple):
        path_tuple = path
    else:
        raise ValueError(f'Value {path} of type {type(path)} is not supported')
    return b''.join([encode_key(key) + _PATH_SENTINEL for key in path_tuple])



import os
import subprocess
import sys

from aim.web.configs import AIM_ENV_MODE_KEY


def create_migration_cmd():
    from aim import storage
    storage_dir = os.path.dirname(storage.__file__)
    if os.getenv(AIM_ENV_MODE_KEY, 'prod') == 'prod':
        migrations_config = os.path.join(storage_dir, 'migrations', 'alembic.ini')
    else:
        migrations_config = os.path.join(storage_dir, 'migrations', 'alembic_dev.ini')
    return [sys.executable, '-m', 'alembic', '-c', migrations_config, 'upgrade', 'head']


def upgrade_database(db_url):
    migration_env = os.environ.copy()
    migration_env['AIM_RUN_META_DATA_DB_URL'] = db_url

    cmd = create_migration_cmd()

    migration_command = subprocess.Popen(cmd, env=migration_env, universal_newlines=True)
    migration_command.communicate()
    exit_code = migration_command.wait()
    if exit_code != 0:
        raise subprocess.SubprocessError(f'Database upgrade failed with exit code {exit_code}')




import logging

import datetime
import json
import pytz
import sys

from collections import defaultdict
from functools import partialmethod
from inspect import currentframe, getframeinfo

from aim.sdk.base_run import BaseRun
from aim.sdk.sequence import Sequence
from aim.sdk.tracker import RunTracker
from aim.sdk.reporter import RunStatusReporter, ScheduledStatusReporter
from aim.sdk.run_heartbeat_reporter import RemoteRunHeartbeatReporter
from aim.sdk.sequence_collection import SingleRunSequenceCollection
from aim.sdk.utils import (
    backup_run,
)

from aim.sdk.types import AimObject
from aim.sdk.logging import LogRecord, LogRecords

from aim.storage.treeview import TreeView
from aim.storage.context import Context
from aim.storage import treeutils

from aim.ext.resource import ResourceTracker, DEFAULT_SYSTEM_TRACKING_INT
from aim.ext.cleanup import AutoClean
from aim.ext.utils import (
    get_environment_variables,
    get_installed_packages,
    get_git_info,
)

from typing import Any, Dict, Iterator, Optional, Tuple, Union
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pandas import DataFrame

    from aim.sdk.sequences.metric import Metric
    from aim.sdk.sequences.image_sequence import Images
    from aim.sdk.sequences.audio_sequence import Audios
    from aim.sdk.sequences.distribution_sequence import Distributions
    from aim.sdk.sequences.figure_sequence import Figures
    from aim.sdk.sequences.text_sequence import Texts
    from aim.sdk.sequence_collection import SequenceCollection
    from aim.ext.resource.log import Logs
    from aim.sdk.repo import Repo

logger = logging.getLogger(__name__)


class BasicRunAutoClean(AutoClean['BasicRun']):
    PRIORITY = 90

    def __init__(self, instance: 'Run') -> None:
        """
        Prepare the `Run` for automatic cleanup.

        Args:
            instance: The `Run` instance to be cleaned up.
        """
        super().__init__(instance)

        self.read_only = instance.read_only
        self.hash = instance.hash
        self.meta_run_tree = instance.meta_run_tree
        self.repo = instance.repo

        self.extra_resources = []

        self._tracker = instance._tracker
        self._checkins = instance._checkins
        self._heartbeat = instance._heartbeat
        self._lock = instance._lock

    def add_extra_resource(self, resource) -> None:
        self.extra_resources.append(resource)

    def finalize_run(self):
        """
        Finalize the run by indexing all the data.
        """
        self.meta_run_tree['end_time'] = datetime.datetime.now(pytz.utc).timestamp()

    def empty_rpc_queue(self):
        if self.repo.is_remote_repo:
            self.repo._client.get_queue(self.hash).wait_for_finish()

    def finalize_rpc_queue(self):
        if self.repo.is_remote_repo:
            self.repo._client.get_queue(self.hash).stop()
            self.repo._client.remove_queue(self.hash)

    def _close(self) -> None:
        """
        Close the `Run` instance resources and trigger indexing.
        """
        if self.read_only:
            logger.debug(f'Run {self.hash} is read-only, skipping cleanup')
            return
        for res in reversed(self.extra_resources):
            logger.debug(f'Closing resource {res}')
            res.close()

        self.empty_rpc_queue()
        self.finalize_run()
        if self._heartbeat is not None:
            self._heartbeat.stop()
        if self._checkins is not None:
            self._checkins.close()
        if self._lock:
            self._lock.release()
        self.finalize_rpc_queue()


# TODO: [AT] generate automatically based on ModelMappedRun
class StructuredRunMixin:
    @property
    def name(self):
        """Run name, set by user.

            :getter: Returns run's name.
            :setter: Sets run's name.
            :type: string
        """
        return self.props.name

    @name.setter
    def name(self, value):
        self.props.name = value

    @property
    def description(self):
        """Run description, set by user.

            :getter: Returns run's description.
            :setter: Sets run's description.
            :type: string
        """
        return self.props.description

    @description.setter
    def description(self, value):
        self.props.description = value

    @property
    def archived(self):
        """Check is run archived or not.

            :getter: Returns run's archived state.
            :setter: Archive/un-archive run.
            :type: bool
        """
        return self.props.archived

    @archived.setter
    def archived(self, value):
        self.props.archived = value

    @property
    def creation_time(self):
        """Run object creation time [UTC] as timestamp.

            :getter: Returns run creation time.
        """
        return self.props.creation_time

    @property
    def created_at(self):
        if self.repo.is_remote_repo:
            return datetime.datetime.fromtimestamp(self.creation_time, tz=pytz.utc).replace(tzinfo=None)
        else:
            return self.props.created_at

    @property
    def finalized_at(self):
        if self.end_time:
            return datetime.datetime.fromtimestamp(self.end_time, tz=pytz.utc).replace(tzinfo=None)
        else:
            return None

    @property
    def end_time(self):
        """Run finalization time [UTC] as timestamp.

            :getter: Returns run finalization time.
        """
        try:
            return self.meta_run_tree['end_time']
        except KeyError:
            # run saved with old version. fallback to sqlite data
            return self.props.end_time

    @property
    def duration(self):
        """Run duration in seconds (end_time - creation_time)

            :getter: Returns run duration.
        """
        if self.end_time:
            return self.end_time - self.creation_time
        else:
            from aim.web.api.utils import datetime_now
            return datetime_now().timestamp() - self.creation_time

    @property
    def active(self):
        """Check if run is active or not.

            :getter: Returns run's active state.
            :type: bool
        """

        if self.end_time:
            return False
        else:
            return True

    @property
    def experiment(self):
        """Run experiment.

            :getter: Returns run's experiment name.
            :setter: Sets run's experiment.
            :type: string
        """
        return self.props.experiment

    @experiment.setter
    def experiment(self, value):
        self.props.experiment = value

    @property
    def tags(self):
        """List of run tags.

            :getter: Returns run's tag list.
        """
        return self.props.tags

    def add_tag(self, value):
        """Add tag to run

        Args:
            value (str): Tag to add.
        """
        return self.props.add_tag(value)

    def remove_tag(self, tag_name):
        """Remove run tag.

        Args:
            tag_name (str): :obj:`name` of tag to be removed.
        """
        return self.props.remove_tag(tag_name)


class BasicRun(BaseRun, StructuredRunMixin):

    _metric_version_warning_shown = False

    def __init__(self, run_hash: Optional[str] = None, *,
                 repo: Optional[Union[str, 'Repo']] = None,
                 read_only: bool = False,
                 experiment: Optional[str] = None,
                 force_resume: bool = False,
                 ):
        self._resources: Optional[BasicRunAutoClean] = None
        super().__init__(run_hash, repo=repo, read_only=read_only)

        self.meta_attrs_tree: TreeView = self.meta_tree.subtree('attrs')
        self.meta_run_attrs_tree: TreeView = self.meta_run_tree.subtree('attrs')

        if not read_only:
            logger.debug(f'Opening Run {self.hash} in write mode')

            if self.check_metrics_version():
                if self.repo.is_remote_repo:
                    logger.warning(f'Cannot track Run with remote repo {self.repo.path}. Please upgrade repo first '
                                   f'with the following command:')
                    logger.warning(f'aim storage --repo {self.repo.path} upgrade 3.11+ \'*\'')
                    raise RuntimeError
                else:
                    logger.warning(f'Detected sub-optimal format metrics for Run {self.hash}. Upgrading...')
                    backup_path = backup_run(self)
                    try:
                        self.update_metrics()
                        logger.warning(f'Successfully converted Run {self.hash}')
                        logger.warning(f'Run backup can be found at {backup_path}. '
                                       f'In case of any issues the following command can be used to restore data: '
                                       f'`aim storage --repo {self.repo.root_path} restore {self.hash}`')
                    except Exception as e:
                        logger.error(f'Failed to convert metrics. {e}')
                        logger.warning(f'Run backup can be found at {backup_path}. '
                                       f'To restore data please run the following command: '
                                       f'`aim storage --repo {self.repo.root_path} restore {self.hash}`')
                        raise

        self._props = None
        self._checkins = None
        self._heartbeat = None

        if not read_only:
            if not self.repo.is_remote_repo:
                self._checkins = RunStatusReporter(self.hash, self.repo.path)
                self._heartbeat = ScheduledStatusReporter(self._checkins)
            else:
                self._heartbeat = RemoteRunHeartbeatReporter(self.repo._client, self.hash)

            try:
                self.meta_run_attrs_tree.first_key()
            except (KeyError, StopIteration):
                # no run params are set. use empty dict
                self[...] = {}
            self.meta_run_tree['end_time'] = None
            self.props
        if experiment:
            self.experiment = experiment

        self._tracker = RunTracker(self)
        self._resources = BasicRunAutoClean(self)

    def __hash__(self) -> int:
        return super().__hash__()

    def idx_to_ctx(self, idx: int) -> Context:
        return self._tracker.idx_to_ctx(idx)

    def __setitem__(self, key: str, val: Any):
        """Set Run top-level meta-parameter.

        Args:
             key (:obj:`str`): Top-level meta-parameter name. Use ellipsis to reset
                run's all meta-parameters.
             val: Meta-parameter value.

        Examples:
            >>> run = Run('3df703c')
            >>> run[...] = params
            >>> run['hparams'] = {'batch_size': 42}
        """
        self.meta_run_attrs_tree[key] = val
        self.meta_attrs_tree[key] = val

    def __getitem__(self, key):
        """Get run meta-parameter by key.

        Args:
            key: path to Run meta-parameter.
        Returns:
            Collected sub-tree of Run meta-parameters.
        Examples:
            >>> run = Run('3df703c')
            >>> run['hparams']  # -> {'batch_size': 42}
            >>> run['hparams', 'batch_size']  # -> 42
        """
        return self._collect(key)

    def set(self, key, val: Any, strict: bool = True):
        self.meta_run_attrs_tree.set(key, val, strict)
        self.meta_attrs_tree.set(key, val, strict)

    def get(self, key, default: Any = None, strict: bool = True, resolve_objects=False):
        try:
            return self._collect(key, strict=strict, resolve_objects=resolve_objects)
        except KeyError:
            return default

    def _collect(self, key, strict: bool = True, resolve_objects: bool = False):
        return self.meta_run_attrs_tree.collect(key, strict=strict, resolve_objects=resolve_objects)

    def __delitem__(self, key: str):
        """Remove key from run meta-params.
        Args:
            key: meta-parameter path
        """
        del self.meta_attrs_tree[key]
        del self.meta_run_attrs_tree[key]

    def track(
        self,
        value,
        name: str = None,
        step: int = None,
        epoch: int = None,
        *,
        context: AimObject = None,
    ):
        """Main method for tracking numeric value series and object series.

        Args:
             value: The tracked value.
             name (str): Tracked sequence name.
             step (:obj:`int`, optional): Sequence tracking iteration. Auto-incremented if not specified.
             epoch (:obj:`int`, optional): The training epoch.
             context (:obj:`dict`, optional): Sequence tracking context.

        Appends the tracked value to sequence specified by `name` and `context`.
        Appended values should be of the same type, in other words, sequence is a homogeneous collection.
        """

        self._tracker(value, name, step, epoch, context=context)

    # logging API
    def _log_message(self, level: int, msg: str, **params):
        frame_info = getframeinfo(currentframe().f_back)
        logger_info = (frame_info.filename, frame_info.lineno)
        self.track(LogRecord(msg, level, logger_info=logger_info, **params), name='__log_records')
        block = (level > logging.WARNING)
        self._checkins.check_in(flag_name="new_logs", block=block)

    log_error = partialmethod(_log_message, logging.ERROR)
    log_warning = partialmethod(_log_message, logging.WARNING)
    log_info = partialmethod(_log_message, logging.INFO)
    log_debug = partialmethod(_log_message, logging.DEBUG)

    def get_log_records(self) -> Optional[LogRecords]:
        """Retrieve duplicated terminal logs for a run

        Returns:
            :obj:`Sequence` object if exists, `None` otherwise.
        """
        return self._get_sequence('log_records', '__log_records', Context({}))

    @property
    def props(self):
        if self._props is None:
            self._props = self.repo.request_props(self.hash, self.read_only)
        return self._props

    def iter_metrics_info(self) -> Iterator[Tuple[str, Context, 'Run']]:
        """Iterator for all run metrics info.

        Yields:
            tuples of (name, context, run) where run is the Run object itself and
            name, context defines Metric type sequence (with values of `float` and `int`).
        """
        yield from self.iter_sequence_info_by_type(('float', 'int'))

    def iter_sequence_info_by_type(self, dtypes: Union[str, Tuple[str, ...]]) -> Iterator[Tuple[str, Context, 'Run']]:
        """Iterator for run sequence infos for the given object data types

        Args:
             dtypes: The objects data types list.

        Yields:
            tuples of (name, context, run) where run is the Run object itself and name, context defines sequence for
            one of `dtypes` types.
        """
        if isinstance(dtypes, str):
            dtypes = (dtypes,)
        for ctx_idx, run_ctx_dict in self.meta_run_tree.subtree('traces').items():
            assert isinstance(ctx_idx, int)
            ctx = self.idx_to_ctx(ctx_idx)
            # run_ctx_view = run_meta_traces.view(ctx_idx)
            for seq_name in run_ctx_dict.keys():
                assert isinstance(seq_name, str)
                # skip sequences not matching dtypes.
                # sequences with no dtype are considered to be float sequences.
                # '*' stands for all data types
                if '*' in dtypes or run_ctx_dict[seq_name].get('dtype', 'float') in dtypes:
                    yield seq_name, ctx, self

    def metrics(self) -> 'SequenceCollection':
        """Get iterable object for all run tracked metrics.

        Returns:
            :obj:`MetricCollection`: Iterable for run metrics.

        Examples:
            >>> run = Run('3df703c')
            >>> for metric in run.metrics():
            >>>     metric.values.sparse_numpy()
        """
        from aim.sdk.sequences.metric import Metric
        self.repo._prepare_runs_cache()
        return SingleRunSequenceCollection(self, seq_cls=Metric)

    def __eq__(self, other: 'Run') -> bool:
        return self.hash == other.hash and self.repo == other.repo

    def get_metric(
            self,
            name: str,
            context: Context
    ) -> Optional['Metric']:
        """Retrieve metric sequence by it's name and context.

        Args:
             name (str): Tracked metric name.
             context (:obj:`Context`): Tracking context.

        Returns:
            :obj:`Metric` object if exists, `None` otherwise.
        """
        if self.read_only and not Run._metric_version_warning_shown:
            if self.check_metrics_version():
                logger.warning(f'Detected sub-optimal format metrics for Run {self.hash}. Consider upgrading repo '
                               f'to improve queries performance:')
                logger.warning(f'aim storage --repo {self.repo.path} upgrade 3.11+ \'*\'')
                Run._metric_version_warning_shown = True

        return self._get_sequence('metric', name, context)

    def get_image_sequence(
            self,
            name: str,
            context: Context
    ) -> Optional['Images']:
        """Retrieve images sequence by it's name and context.

        Args:
             name (str): Tracked image sequence name.
             context (:obj:`Context`): Tracking context.

        Returns:
            :obj:`Images` object if exists, `None` otherwise.
        """
        return self._get_sequence('images', name, context)

    def get_figure_sequence(
            self,
            name: str,
            context: Context
    ) -> Optional['Figures']:
        """Retrieve figure sequence by its name and context.

        Args:
             name (str): Tracked figure sequence name.
             context (:obj:`Context`): Tracking context.

        Returns:
            :obj:`Figures` object if exists, `None` otherwise.
        """
        return self._get_sequence('figures', name, context)

    def get_audio_sequence(
            self,
            name: str,
            context: Context
    ) -> Optional['Audios']:
        """Retrieve audios sequence by its name and context.

        Args:
             name (str): Tracked audios sequence name.
             context (:obj:`Context`): Tracking context.

        Returns:
            :obj:`Audios` object if exists, `None` otherwise.
        """
        return self._get_sequence('audios', name, context)

    def get_distribution_sequence(
            self,
            name: str,
            context: Context
    ) -> Optional['Distributions']:
        """Retrieve distributions sequence by it's name and context.

        Args:
             name (str): Tracked distribution sequence name.
             context (:obj:`Context`): Tracking context.

        Returns:
            :obj:`Distributions` object if exists, `None` otherwise.
        """
        return self._get_sequence('distributions', name, context)

    def get_terminal_logs(self) -> Optional['Logs']:
        """Retrieve duplicated terminal logs for a run

                Returns:
                    :obj:`Logs` object if exists, `None` otherwise.
                """
        return self._get_sequence('logs', 'logs', Context({}))

    def get_text_sequence(
            self,
            name: str,
            context: Context
    ) -> Optional['Texts']:
        """Retrieve texts sequence by it's name and context.

        Args:
             name (str): Tracked text sequence name.
             context (:obj:`Context`): Tracking context.

        Returns:
            :obj:`Texts` object if exists, `None` otherwise.
        """
        return self._get_sequence('texts', name, context)

    def _get_sequence_dtype(
            self,
            sequence_name: str,
            context: Context
    ) -> str:
        try:
            return self.meta_run_tree.subtree(('traces', hash(context), sequence_name, 'dtype')).collect()
        except KeyError:
            # fallback to `float`, cause in older versions there was no `dtype`
            return 'float'

    def _get_sequence(
            self,
            seq_type: str,
            sequence_name: str,
            context: Context
    ) -> Optional[Sequence]:
        seq_cls = Sequence.registry.get(seq_type, None)
        if seq_cls is None:
            raise ValueError(f'\'{seq_type}\' is not a valid Sequence')
        assert issubclass(seq_cls, Sequence)
        tracked_dtype = self._get_sequence_dtype(sequence_name, context)
        if tracked_dtype not in seq_cls.allowed_dtypes():
            return None
        sequence = seq_cls(sequence_name, context, self)
        return sequence if bool(sequence) else None

    def collect_sequence_info(
            self,
            sequence_types: Union[str, Tuple[str, ...]],
            skip_last_value=False
    ) -> Dict[str, list]:
        """Retrieve Run's all sequences general overview.

        Args:
             sequence_types: Type names of sequences for which to collect name/context pairs.
             skip_last_value (:obj:`bool`, optional): Boolean flag to include tracked sequence last value in
             sequence info. False by default.

        Returns:
             :obj:`list`: list of sequence's `context`, `name` and optionally last tracked value triplets.
        """
        traces = self.meta_run_tree.subtree('traces')
        traces_overview = {}

        # build reverse map of sequence supported dtypes
        dtype_to_sequence_type_map = defaultdict(list)
        if isinstance(sequence_types, str):
            sequence_types = (sequence_types,)
        for seq_type in sequence_types:
            traces_overview[seq_type] = []
            seq_cls = Sequence.registry.get(seq_type, None)
            if seq_cls is None:
                raise ValueError(f'\'{seq_type}\' is not a valid Sequence')
            assert issubclass(seq_cls, Sequence)
            dtypes = seq_cls.allowed_dtypes()
            for dtype in dtypes:
                dtype_to_sequence_type_map[dtype].append(seq_type)

        for idx in traces.keys():
            ctx_dict = self.idx_to_ctx(idx).to_dict()
            for name, value in traces[idx].items():
                dtype = value.get('dtype', 'float')  # old sequences without dtype set are considered float sequences
                if dtype in dtype_to_sequence_type_map:
                    trace_data = {
                        'context': ctx_dict,
                        'name': name,
                    }
                    if not skip_last_value:
                        trace_data['last_value'] = value
                    for seq_type in dtype_to_sequence_type_map[dtype]:
                        traces_overview[seq_type].append(trace_data)
        return traces_overview

    def _cleanup_trees(self):
        del self.meta_run_attrs_tree
        del self.meta_attrs_tree
        del self.meta_run_tree
        del self.meta_tree
        del self._series_run_trees
        self.meta_run_attrs_tree = None
        self.meta_run_tree = None
        self.meta_attrs_tree = None
        self.meta_tree = None
        self._series_run_trees = None

    def close(self):
        if self._resources is None:
            return
        self._resources.close()
        self._tracker.sequence_infos.clear()
        # de-reference trees and other resources
        del self._resources
        del self._props
        self._resources = None
        self._props = None
        self._cleanup_trees()

    def finalize(self):
        if self._resources is None:
            return

        self._resources.finalize_run()

    def dataframe(
            self,
            include_props: bool = True,
            include_params: bool = True,
    ) -> 'DataFrame':
        """Get run properties and params as pandas DataFrame

        Args:
             include_props: (:obj:`int`, optional): If true, include run structured props
             include_params: (:obj:`int`, optional): If true, include run parameters
        """
        data = {
            'hash': self.hash,
        }

        if include_props:
            # TODO [GA]: Auto collect props based on StructuredRunMixin:
            #  - Exclude created_at, updated_at, finalized_at auto-populated fields
            #  - Collect list of representations in case of ModelMappedCollection's
            data['name'] = self.props.name
            data['description'] = self.props.description
            data['archived'] = self.props.archived
            data['creation_time'] = self.props.creation_time
            data['end_time'] = self.end_time
            data['active'] = self.active
            data['experiment'] = self.props.experiment
            data['tags'] = json.dumps(self.props.tags)

        if include_params:
            # TODO [GA]:
            #  - Move run params collection to utility function
            #  - Remove code duplication from Metric.dataframe
            for path, val in treeutils.unfold_tree(self[...],
                                                   unfold_array=False,
                                                   depth=3):
                s = ''
                for key in path:
                    if isinstance(key, str):
                        s += f'.{key}' if len(s) else f'{key}'
                    else:
                        s += f'[{key}]'

                if isinstance(val, (tuple, list, dict)):
                    val = json.dumps(val)
                if s not in data.keys():
                    data[s] = val

        import pandas as pd
        df = pd.DataFrame(data, index=[0])
        return df

    def report_progress(
        self,
        *,
        expect_next_in: int = 0,
        block: bool = False,
    ) -> None:
        """
        Report progress for the run. Report the expected time for the next progress report.

        If no progress reports are received by the expiry date (plus the grace period), the
        run is considered as failed.

        Args:
            expect_next_in: (:obj:`int`, optional): The number of seconds to wait before the next progress report.
            block: (:obj:`bool`, optional): If true, block the thread until the report is written to filesystem.
        """
        if self._checkins is None:
            raise ValueError('Progress reports are not enabled for this run')
        self._checkins._check_in(expect_next_in=expect_next_in, block=block)

    def report_successful_finish(
        self,
        *,
        block: bool = True,
    ) -> None:
        """
        Report successful finish of the run. If the run is not marked as successfully finished,
        it can potentially be considered as failed.

        Args:
            block: (:obj:`bool`, optional): If true, block the thread until the report is written to filesystem.
        """
        if self._checkins is None:
            raise ValueError('Progress reports are not enabled for this run')
        self._checkins._report_successful_finish(block=block)


class Run(BasicRun):
    """Run object used for tracking metrics.

    Provides method :obj:`track` to track value and object series for multiple names and contexts.
    Provides dictionary-like interface for Run object meta-parameters.
    Provides API for iterating through tracked sequences.

    Args:
         run_hash (:obj:`str`, optional): Run's hash. If skipped, generated automatically.
         repo (:obj:`Union[Repo,str], optional): Aim repository path or Repo object to which Run object is bound.
            If skipped, default Repo is used.
         read_only (:obj:`bool`, optional): Run creation mode.
            Default is False, meaning Run object can be used to track metrics.
         experiment (:obj:`str`, optional): Sets Run's `experiment` property. 'default' if not specified.
            Can be used later to query runs/sequences.
         force_resume (:obj:`bool`, optional): Forcefully resume stalled Run.
         system_tracking_interval (:obj:`int`, optional): Sets the tracking interval in seconds for system usage
            metrics (CPU, Memory, etc.). Set to `None` to disable system metrics tracking.
         log_system_params (:obj:`bool`, optional): Enable/Disable logging of system params such as installed packages,
            git info, environment variables, etc.
    """

    def __init__(self, run_hash: Optional[str] = None, *,
                 repo: Optional[Union[str, 'Repo']] = None,
                 read_only: bool = False,
                 experiment: Optional[str] = None,
                 force_resume: bool = False,
                 system_tracking_interval: Optional[Union[int, float]] = DEFAULT_SYSTEM_TRACKING_INT,
                 log_system_params: Optional[bool] = False,
                 capture_terminal_logs: Optional[bool] = True):
        super().__init__(run_hash, repo=repo, read_only=read_only, experiment=experiment, force_resume=force_resume)

        self._system_resource_tracker: ResourceTracker = None
        if not read_only:
            if log_system_params:
                self['__system_params'] = {
                    'packages': get_installed_packages(),
                    'env_variables': get_environment_variables(),
                    'git_info': get_git_info(),
                    'executable': sys.executable,
                    'arguments': sys.argv
                }

            if ResourceTracker.check_interval(system_tracking_interval) or capture_terminal_logs:
                current_logs = self.get_terminal_logs()
                log_offset = current_logs.last_step() + 1 if current_logs else 0
                self._system_resource_tracker = ResourceTracker(self._tracker,
                                                                system_tracking_interval,
                                                                capture_terminal_logs,
                                                                log_offset)
                self._system_resource_tracker.start()
                self._resources.add_extra_resource(self._system_resource_tracker)
